"""Method pointers and config for constraints"""

from .input import (
    Linear_Constraint,
    BatchVariance_Constraint
)

from .output import (
    Blank_Constraint,
    L1_Constraint
)

const_class_dict = {'Linear_Constraint': Linear_Constraint,
                    'BatchVariance_Constraint': BatchVariance_Constraint,
                    'Blank_Constraint': Blank_Constraint,
                    'L1_Constraint': L1_Constraint
                    }



"""Constraints on the input features of a parameter space"""

from obsidian.parameters import IParamSpace, Param_Continuous
from obsidian.constraints import Constraint
from obsidian.config import TORCH_DTYPE

import torch
from torch import Tensor
from abc import abstractmethod

# See  https://botorch.org/api/optim.html, optimize_acqf "equality_constraints", "inequality_constraints",
# and "nonlinear_inequality_constraints"


class Input_Constraint(Constraint):
    """
    Input constraint for a given parameter space.

    Note: Saving and loading input constraints is managed by ParamSpace.
        The interface class IParamSpace is used here to avoid circular imports
        with constraints that depend on ParamSpace, which saves/loads constraints.
    """
    def __init__(self,
                 X_space: IParamSpace):
        super().__init__()
        self.X_space = X_space


class Linear_Constraint(Input_Constraint):
    """
    Input constraint for a given parameter space.

    Note: SUM(LHS) <= RHS equivalent to -SUM(LHS) >= -RHS

    Linear constraints must return:
        tuple = (indices = Tensor, coefficients = Tensor, rhs = float)
        where sum_i(X[indices[i]] * coefficients[i]) = rhs (equality) or >= rhs (inequality)

    Attributes:
        X_space (ParamSpace): The parameter space object.
        ind (list[float | int], optional): The indices of the parameters to be constrained.
            Defaults to ``[0]``.
        weights (list[float | int | list], optional): The coefficients for the parameters in the
            constraint equation. Defaults to ``[1]``.
        rhs (int | float, optional): The right-hand side value of the constraint equation.
            Defaults to ``-1``.
        equality (bool, optional): Whether the constraint is an equality (=)  or inequality (>=)
            constraint. Defaults to ``False`` for inequality constraint.

    """
    def __init__(self,
                 X_space: IParamSpace,
                 ind: list[float | int] = [0],
                 weights: list[float | int] = [1],
                 rhs: int | float = -1,
                 equality: bool = False) -> None:
        super().__init__(X_space)

        self.register_buffer('ind', torch.tensor(ind, dtype=torch.int64))
        self.register_buffer('weights', torch.tensor(weights, dtype=TORCH_DTYPE))
        self.register_buffer('rhs', torch.tensor(rhs, dtype=TORCH_DTYPE))
        self.register_buffer('equality', torch.tensor(equality, dtype=torch.bool))

        # Determine which parameters in the decoded space are indicated
        self.ind_t = torch.tensor([X_space.t_map[i] for i in ind],
                                  dtype=torch.int64)
        for i in ind:
            if not isinstance(X_space.params[i], Param_Continuous):
                raise TypeError('Indeces for input constraints must be \
                                for continuous parameters only')

        # We neeed to write the constraints on the encoded space
        # To convert (raw -> encoded) for a continuous param with min-max scaling:
        #      w_t = w * range
        #      rhs_t = rhs - sum(w_t * min)
        weights_t = []
        rhs_t = rhs
        for w, i in zip(weights, ind):
            param_i = X_space.params[i]
            rhs_t -= param_i.min*w
            weights_t.append(w * param_i.range)
        self.weights_t = torch.tensor(weights_t, dtype=TORCH_DTYPE)
        self.rhs_t = torch.tensor(rhs_t, dtype=TORCH_DTYPE)

    def forward(self) -> tuple[Tensor, Tensor, Tensor]:
        """
        Forward method for the input constraint.

        Forms the linear constraint in the tuple that can be handled by BoTorch optimizer.

        Returns:
            tuple[Tensor, Tensor, Tensor]: A tuple containing the indices, coefficients,
                and right-hand side value of the constraint in the encoded space.

        """
        return self.ind_t, self.weights_t, self.rhs_t

    def __repr__(self) -> str:
        """String representation of object"""
        return (f'{self.__class__.__name__}(ind={self.ind},'
                f'weights={self.weights}, rhs={self.rhs}, equality={self.equality})')


class Nonlinear_Constraint(Input_Constraint):
    """
    Abstract class for nonlinear constraints

    Nonlinear inequality constraints must return:
        tuple = (constraint(x) = callable, intra-point = bool)

        where:
            Intra point: callable takes X in (d) and return scalar
            Intra point: when q = 1, or when applying the same constraint to each
                candidate in the batch
            Inter point: callable takes X in (q x d) and returns scalar
            Inter point: When q > 1 and the constraints are applied to an entire
                batch of candidates
        
        Use 1-d tensor of indices for intra point constraint
        Use 2-d tensor of indices for inter point constraint where indices[i] = (k_i, l_i, ...)

    """
    @abstractmethod
    def forward(self) -> tuple[callable, bool]:
        """
        Forward method for the input constraint.

        Returns:
            tuple[callable, bool]: A tuple containing the callable function and a boolean
                indicating whether the constraint is an intra-point constraint.

        """
        pass  # pragma: no cover


class BatchVariance_Constraint(Nonlinear_Constraint):
    """
    Constraint how much one parameter can vary over a batch of optimized points
    
    Attributes:
        X_space (ParamSpace): The parameter space object.
        indices (list[float | int], optional): The indices of the parameters to be constrained.
            Defaults to ``[0]``.
        coeff (list[float | int | list], optional): The coefficients for the parameters in the
            constraint equation. Defaults to ``[1]``.
        rhs (int | float, optional): The right-hand side value of the constraint equation.
            Defaults to ``-1``.

    """
    def __init__(self,
                 X_space: IParamSpace,
                 ind: int,
                 tol: int | float = 0.01) -> None:
        
        super().__init__(X_space)

        self.register_buffer('ind', torch.tensor(ind, dtype=torch.int64))
        self.register_buffer('tol', torch.tensor(tol, dtype=TORCH_DTYPE))

        # Determine which parameters in the decoded space are indicated
        self.ind_t = torch.tensor(X_space.t_map[ind], dtype=torch.int64)
        if not isinstance(X_space.params[ind], Param_Continuous):
            raise TypeError('Indeces for input constraints must be \
                            for continuous parameters only')

    def forward(self):

        def nl_func(X: Tensor) -> Tensor:
            X_range = X.max(dim=0).values - X.min(dim=0).values
            target_range = X_range[self.ind_t]
            return self.tol - target_range
        
        nonlinear_constraint = (nl_func, False)  # False for inter-point

        return nonlinear_constraint
    
    def __repr__(self) -> str:
        """String representation of object"""
        return (f'{self.__class__.__name__}(ind={self.ind},'
                f'tol={self.tol})')



"""Plotly figure-generating functions"""

from obsidian.campaign import Campaign
from obsidian.optimizer import Optimizer
from obsidian.exceptions import UnfitError, UnsupportedError
from obsidian.parameters import Param_Continuous
from .branding import obsidian_colors
from obsidian.plotting.branding import obsidian_color_list as colors

import plotly.graph_objects as go
from plotly.graph_objects import Figure
from plotly.subplots import make_subplots

import pandas as pd
import numpy as np
import math


def visualize_inputs(campaign: Campaign) -> Figure:
    """
    Visualizes the input variables of a campaign.

    Args:
        campaign (Campaign): The campaign object containing the input data.

    Returns:
        Figure: The plotly Figure object containing the visualization.
    """
    n_dim = campaign.X_space.n_dim
    X = campaign.X
    
    # Enforce that there are 2 rows
    # Determine the number of columns based on the number of dimensions
    rows = 2
    cols = math.ceil(n_dim / rows)
    
    height = 200 * rows
    width = 300 * cols
    fontsize = 8

    color_list = colors * 10
    
    # Add an extra 2 cols for the correlation matrix
    fig = make_subplots(
        rows=rows, cols=cols + 2,
        vertical_spacing=0.2,
        horizontal_spacing=0.1,
        specs=[[{}]*cols + [{"colspan": 2, "rowspan": 2}, None],
               [{}]*cols + [None, None]],
        subplot_titles=[X.columns[i] for i in range(cols)]
        + ['Correlation Matrix']
        + [X.columns[i] for i in range(cols, n_dim)]
    )
    # if campaign.optimizer is fitted, then X_best_f_idx is identified
    if 'X_best_f_idx' in dir(campaign.optimizer):
        marker_shapes = ['diamond' if rowInd in [campaign.optimizer.X_best_f_idx] else 'circle' for rowInd in range(campaign.X.shape[0])]
    else:
        marker_shapes = ['circle']*campaign.X.shape[0]
    
    for i, param in enumerate(X.columns):
        row_i = i // cols + 1
        col_i = i % cols + 1
        fig.add_trace(go.Scatter(x=X.index, y=X[param],
                                 mode='markers', name=param,
                                 marker=dict(color=color_list[i], symbol=marker_shapes),
                                 showlegend=False),
                      row=row_i, col=col_i)
        fig.update_xaxes(tickvals=np.around(np.linspace(0, campaign.m_exp, 5)),
                         row=row_i, col=col_i)
    
    # Add note to explain the shape of markers
    if hasattr(campaign.optimizer, 'X_best_f_idx'):
        fig.add_annotation(
            text="Note: The diamond markers denote samples that achieve the best sum of targets.",
            showarrow=False,
            xref="paper", yref="paper",
            x=0,
            y=-0.2,
            font=dict(style="italic")
        )
    
    # Calculate the correlation matrix
    X_u = campaign.X_space.unit_map(X)
    corr_matrix = X_u.corr()
    fig.add_trace(go.Heatmap(z=corr_matrix.values,
                             x=corr_matrix.columns,
                             y=corr_matrix.columns,
                             colorscale=[[0, obsidian_colors.rich_blue],
                                         [0.5, obsidian_colors.teal],
                                         [1, obsidian_colors.lemon]],
                             name='Correlation'),
                  row=1, col=cols+1)
    
    fig.update_yaxes(showticklabels=False, row=1, col=cols+1)
    fig.update_xaxes(tickangle=-90, row=1, col=cols+1)
    
    fig.update_layout(width=width, height=height, template='ggplot2',
                      font_size=fontsize, title_text='Campaign Data Visualization')
    fig.update_annotations(font_size=fontsize)
    
    return fig


def MDS_plot(campaign: Campaign) -> Figure:
    """
    Creates a Multi-Dimensional Scaling (MDS) plot of the campaign data,
    colored by iteration.
    
    This plot is helpful to visualize the convergence of the optimizer on a 2D plane.
    
    Args:
        campaign (Campaign): The campaign object containing the data.
        
    Returns:
        fig (Figure): The MDS plot
    """
    try:
        from sklearn.manifold import MDS
    except ImportError:
        raise ImportError('The `sklearn` package (>1.0) is required for the MDS plot. \
                          Please install it using `pip install scikit-learn`')

    mds = MDS(n_components=2)
    X_mds = mds.fit_transform(campaign.X_space.encode(campaign.X))

    iter_max = campaign.data['Iteration'].max()
    iter_vals = campaign.data['Iteration'].values
    
    if campaign.data['Iteration'].nunique() == 1:
        iter_vals = np.zeros_like(iter_vals)
        iter_max = 0
        cbar = None
    else:
        cbar = dict(title=dict(text='Iteration', font=dict(size=10)))
    
    fig = go.Figure()

    fig.add_trace(go.Scatter(x=X_mds[:, 0], y=X_mds[:, 1],
                             mode='markers',
                             name='',
                             marker={'color': iter_vals, 'size': 10,
                                     'cmax': iter_max, 'cmin': 0,
                                     'colorscale': [[0, obsidian_colors.rich_blue],
                                                    [0.5, obsidian_colors.teal],
                                                    [1, obsidian_colors.lemon]],
                                     'colorbar': cbar
                                     },
                             customdata=campaign.data[
                                 list(campaign.X_space.X_names) + ['Iteration']],
                             showlegend=False
                             ))

    template = ["<b>"+str(param.name)+"</b>: "+" %{customdata["+str(i)+"]"
                + (":.3G}"if isinstance(param, Param_Continuous) else "}") + "<br>"
                for i, param in enumerate(campaign.X_space)]
    
    fig.update_traces(hovertemplate=''.join(template)
                      + '<b>Iteration</b>'
                      + ": %{customdata["+str(len(campaign.X_space))+"]}<br>"
                      + '<b>MDS C1</b>' + ": %{x:.3G}<br>"
                      + '<b>MDS C2</b>' + ": %{y:.3G}<br>")
    
    fig.update_xaxes(title_text='Component 1')
    fig.update_yaxes(title_text='Component 2')
    fig.update_layout(template='ggplot2', title='Multi-Dimensional Scaling (MDS) Plot',
                      autosize=False, height=400, width=500)
    
    return fig


def parity_plot(optimizer: Optimizer,
                f_transform: bool = False,
                response_id: int = 0) -> Figure:
    """
    Produces a plot of surrogate model predictions at an OFAT range in one variable.

    Args:
        optimizer (Optimizer): The optimizer object which contains a surrogate that has been fit to data
            and can be used to make predictions.
        f_transform (bool, optional): An indicator for whether or not to plot the response value in the "objective
            function" form which is directly used by the optimizer, else using the "measured response" form which the
            optimizer preproceses. Default value is ``False`` which plots the raw measured response scale.
        response_id (int, optional): Index of the response for potential multi-response models.
            Default value is ``0`` (single-response).

    Returns:
        fig (Figure): The optimizer fit parity plot

    Raises:
        TypeError: If the optimizer is not an instance of obsidian Optimizer
        UnfitError: If the optimizer is not fit
        ValueError: If the response_id is not a valid response index
    """

    if not isinstance(optimizer, Optimizer):
        raise TypeError('Optimizer must be an instance of obsidian Optimizer')
    if not optimizer.is_fit:
        raise UnfitError('Optimizer must be fit before plotting predictions')
    if response_id < 0 or response_id >= len(optimizer.target):
        raise ValueError('response_id must be a valid response index')

    X = optimizer.X_train
    y_pred = optimizer.predict(X, return_f_inv=not f_transform)
    y_true = optimizer.f_train if f_transform else optimizer.y_train

    y_name = optimizer.y_names[response_id]

    y_lb = y_pred[y_name+' lb'].values
    y_ub = y_pred[y_name+' ub'].values
    y_pred = y_pred[y_name+' (pred)'].values
    y_true = y_true[y_name].values
    
    RMSE = ((y_true-y_pred)/y_true)**2
    NRMSE = RMSE/(y_true.max()-y_true.min())
    
    y_min = np.min([y_true.min(), y_pred.min()])
    y_max = np.max([y_true.max(), y_pred.max()])
    abs_margin = 0.1
    y_abs = [y_min/(1+abs_margin), y_max*(1+abs_margin)]
    
    error_y = y_ub - y_pred
    error_y_minus = y_pred - y_lb
    
    fig = go.Figure()

    fig.add_trace(go.Scatter(x=y_true, y=y_pred,
                             error_y={'array': [f'{y:.3G}' for y in error_y],
                                      'arrayminus': [f'{y:.3G}' for y in error_y_minus],
                                      'color': 'gray', 'thickness': 0.5},
                             mode='markers',
                             name='Observations',
                             marker={'color': NRMSE, 'size': 15,
                                     'cmax': 0.5, 'cmin': 0,
                                     'colorscale': [[0, obsidian_colors.rich_blue],
                                                    [0.5, obsidian_colors.teal],
                                                    [1, obsidian_colors.lemon]],
                                     'colorbar': dict(title=dict(text='NRMSE', font=dict(size=10)))
                                     },
                             showlegend=False
                             ))
    
    fig.update_traces(hovertemplate="(%{x:.3G}, %{y:.3G}) +%{error_y.array:.3G}/-%{error_y.arrayminus:.3G}")
    
    fig.add_trace(go.Scatter(x=y_abs, y=y_abs,
                             mode='lines',
                             name='Parity',
                             showlegend=False,
                             line={'color': 'black', 'dash': 'dot'}))

    fig.update_xaxes(title_text=f'Actual Response ({y_name})')
    fig.update_yaxes(title_text=f'Predicted Response ({y_name})')
    fig.update_layout(template='ggplot2', title='Parity Plot',
                      autosize=False, height=400, width=500)
    
    return fig


def factor_plot(optimizer: Optimizer,
                feature_id: int = 0,
                response_id: int = 0,
                f_transform: bool = False,
                X_ref: pd.DataFrame | None = None, plotRef: bool = True,
                ylim: tuple[float, float] | None = None) -> Figure:
    """
    Produces a plot of surrogate model predictions at an OFAT range in one variable.

    Args:
        optimizer (Optimizer): The optimizer object which contains a surrogate that has been fit to data
            and can be used to make predictions.
        feature_id (int, optional): The index of the desired variable to plot from the last data used
            to fit the surrogate model. The default value is ``0``.
        response_id (int, optional): Index of the response for potential multi-response models.
            Default value is ``0`` (single-response).
        f_transform (bool, optional): An indicator for whether or not to plot the response value in the "objective
            function" form which is directly used by the optimizer, else using the
            "measured response" form which the optimizer preproceses. Default value is
            ``False`` which plots the raw measured response scale.
        plotRef (bool, optional): An indicator for whether or not to plot the reference data points.
            Default value is ``True``.
        ylim (tuple, optional): The y-axis limits for the plot. Default value is ``None``.

    Returns:
        fig (Figure): The matplotlib plot of response value versus 1 predictor variable.

    Raises:
        TypeError: If the optimizer is not an instance of obsidian Optimizer
        UnfitError: If the optimizer is not fit
        ValueError: If the feature_id is not a valid feature index
        ValueError: If the response_id is not a valid response index
        ValueError: If X_ref is provided and not a pd.Series
    """
    if not isinstance(optimizer, Optimizer):
        raise TypeError('Optimizer must be an instance of obsidian Optimizer')
    if not optimizer.is_fit:
        raise UnfitError('Optimizer must be fit before plotting predictions')
    if feature_id >= len(optimizer.X_space):
        raise ValueError('feature_id must be a valid feature index')
    if response_id < 0 or response_id >= len(optimizer.target):
        raise ValueError('response_id must be a valid response index')

    # Create a dataframe of test samples for plotting
    n_samples = 100
    if X_ref is None:
        X_ref = optimizer.X_best_f
    else:
        if not isinstance(X_ref, pd.DataFrame):
            raise TypeError('X_ref must be a DataFrame')
    X_test = pd.concat([X_ref]*n_samples, axis=0).reset_index(drop=True)
    
    # Vary the indicated column
    X_name = X_test.columns[feature_id]
    param_i = optimizer.X_space.params[feature_id]
    unit_span = np.linspace(0, 1-1e-15, n_samples)  # Prevent issues with cat[1.0] = undefined
    X_test[X_name] = param_i.unit_demap(unit_span)
    X = X_test[X_name].values
    
    Y_pred = optimizer.predict(X_test, return_f_inv=not f_transform, PI_range=0.95)
    y_name = optimizer.y_names[response_id]

    Y_mu = Y_pred[y_name+('_t (pred)' if f_transform else ' (pred)')].values
    LCB = Y_pred[y_name+('_t lb' if f_transform else ' lb')].values
    UCB = Y_pred[y_name+('_t ub' if f_transform else ' ub')].values
        
    fig = go.Figure()

    fig.add_trace(go.Scatter(x=np.append(X, X[::-1]), y=np.append(UCB, LCB[::-1]),
                             fill='toself',
                             opacity=0.3,
                             line={'color': obsidian_colors.teal},
                             showlegend=True,
                             name='95% Pred Band'),
                  )
    
    fig.add_trace(go.Scatter(x=X, y=Y_mu,
                             mode='lines',
                             line={'color': obsidian_colors.teal},
                             name='Mean'),
                  )
    if plotRef:
        Y_pred_ref = optimizer.predict(X_ref, return_f_inv=not f_transform)
        Y_mu_ref = Y_pred_ref[y_name+('_t (pred)' if f_transform else ' (pred)')].values
        fig.add_trace(go.Scatter(x=X_ref.iloc[:, feature_id].values, y=Y_mu_ref,
                                 mode='markers',
                                 marker=dict(symbol='diamond'),
                                 line={'color': obsidian_colors.teal},
                                 name='Reference'),
                      )
    fig.update_xaxes(title_text=X_name)
    fig.update_yaxes(title_text=y_name)
    fig.update_layout(template='ggplot2', title=f'Factor Effect Plot for {X_name}')
    fig.update_layout(autosize=False, width=600, height=400)
    if ylim is not None:
        fig.update_layout(yaxis_range=ylim)
    
    return fig


def surface_plot(optimizer: Optimizer,
                 feature_ids: list[int, int] = [0, 1],
                 response_id: int = 0,
                 f_transform: bool = False,
                 plot_bands: bool = True,
                 plot_data: bool = False) -> Figure:
    """
    Produces a surface plot of surrogate model predictions over a 2-parameter grid range.

    Args:
        optimizer (Optimizer): The optimizer object which contains a surrogate that has been fit to data
            and can be used to make predictions.
        feature_ids (list, optional): A list of integers containing the indices of the desired variables
            to plot from the last data used to fit the surrogate model. Default value is ``[0,1]``.
        f_transform (bool, optional): An indicator for whether or not to plot the response value in the "objective
            function" form which is directly used by the optimizer, else using the
            "measured response" form which the optimizer preprocesses. Default value is
            ``False`` which plots the raw measured response scale.
        plot_bands (bool, optional): An indicator for whether or not to plot the confidence bands as a wire
            frame around the surface plot. Default is ``True``.
        plot_data (bool, optional): An indicator for whether or not to plot the raw data locations.
            Default is ``False``, as the data z-height can be misleading for >2D data on a 3D plot.
        response_id (int, optional): Index of the response for potential multi-response models.
            Default value is ``0`` (single-response).

    Returns:
        fig (Figure): The matplotlib plot of surfaces over a 2-parameter grid.

    Raises:
        TypeError: If the optimizer is not an instance of obsidian Optimizer
        UnfitError: If the optimizer is not fit
        ValueError: If the feature_ids are not valid feature indices
        ValueError: If the response_id is not a valid response index
    """
    if not isinstance(optimizer, Optimizer):
        raise TypeError('Optimizer must be an instance of obsidian Optimizer')
    for feature_id in feature_ids:
        if feature_id >= len(optimizer.X_space):
            raise ValueError('feature_id must be a valid feature index')
    if not optimizer.is_fit:
        raise UnfitError('Optimizer must be fit before plotting predictions')
    if response_id < 0 or response_id >= len(optimizer.target):
        raise ValueError('response_id must be a valid response index')
    if plot_data and f_transform:
        raise UnsupportedError('Plotting data is not supported for transformed responses')

    # Create a dataframe of test samples for plotting
    n_grid = 100
    X_test = pd.concat([optimizer.X_best_f]*(n_grid**2), axis=0).reset_index(drop=True)
    
    # Create a mesh grid which is necessary for the 3D plot
    X0_name = X_test.columns[feature_ids[0]]
    X1_name = X_test.columns[feature_ids[1]]
    
    # Vary the indicated column
    x_axes = []
    unit_span = np.linspace(0, 1-1e-15, n_grid)

    params = [optimizer.X_space.params[i] for i in feature_ids]
    for param_i in params:
        x_axes.append(np.array(param_i.unit_demap(unit_span)))

    x0_ax, x1_ax = x_axes
    X0, X1 = np.meshgrid(x0_ax, x1_ax)
    X_test[X0_name] = X0.flatten().reshape(-1, 1)
    X_test[X1_name] = X1.flatten().reshape(-1, 1)

    Y_pred = optimizer.predict(X_test, return_f_inv=not f_transform, PI_range=0.95)
    y_name = optimizer.y_names[response_id]

    Y_mu = Y_pred[y_name+('_t (pred)' if f_transform else ' (pred)')].values
    LCB = Y_pred[y_name+('_t lb' if f_transform else ' lb')].values
    UCB = Y_pred[y_name+('_t ub' if f_transform else ' ub')].values
    
    Z_mu = Y_mu.reshape(n_grid, n_grid)
    Z_LCB = LCB.reshape(n_grid, n_grid)
    Z_UCB = UCB.reshape(n_grid, n_grid)
    
    fig = go.Figure()
    fig.add_trace(go.Surface(z=Z_mu, x=x0_ax, y=x1_ax,
                             name='Mean',
                             opacity=0.85))
    
    # Plot the confidence bands only if desired
    if plot_bands:
        for i, (x0, x1, ucb, lcb) in enumerate(zip(X0, X1, Z_UCB, Z_LCB)):
            fig.add_trace(go.Scatter3d(z=ucb, x=x0, y=x1,
                                       mode='lines',
                                       line={'color': '#b0b0b0', 'width': 2},
                                       opacity=0.5,
                                       name='UCB',
                                       legendgroup='UCB',
                                       showlegend=True if i == 0 else False),
                          )
            fig.add_trace(go.Scatter3d(z=lcb, x=x0, y=x1,
                                       mode='lines',
                                       line={'color': '#b0b0b0', 'width': 2},
                                       opacity=0.5,
                                       name='LCB',
                                       legendgroup='LCB',
                                       showlegend=True if i == 0 else False),
                          )
            
    fig.update_layout(legend_orientation='h', template='ggplot2', title=f'Surface Plot for {X0_name} vs {X1_name}',
                      scene=dict(xaxis_title=X0_name, yaxis_title=X1_name, zaxis_title=y_name))
    
    Z_train = optimizer.f_train if f_transform else optimizer.y_train
    
    # Plot the locations of raw data only if desired
    if plot_data:
        # Use a marker size which is larger if closer to the "other" values of the surface
        # Smaller markers mean there are other variables pulling it away from the surface plot
        if len(optimizer.X_space) != 2:
            X_t_train_ex = optimizer.X_t_train.copy().drop(columns=[X0_name, X1_name]).values
            X_t_test_ex = optimizer.X_space.encode(X_test.copy().drop(columns=[X0_name, X1_name]).drop_duplicates()).values

            if X_t_train_ex.shape[0] > 1:
                X_t_dist = np.linalg.norm(X_t_train_ex-X_t_test_ex, ord=2, axis=1)
                X_t_dist_scaled = (X_t_dist - X_t_dist.min())/(X_t_dist.max()-X_t_dist.min())
                dist_scaled = 15-10*X_t_dist_scaled
            else:
                dist_scaled = 10
        else:
            dist_scaled = 10
        
        fig.add_trace(go.Scatter3d(x=optimizer.X_train[X0_name], y=optimizer.X_train[X1_name],
                                   z=Z_train[y_name],
                                   mode='markers',
                                   marker={'color': '#000000', 'size': dist_scaled},
                                   name='Observations'))

    return fig


def optim_progress(campaign: Campaign,
                   response_ids: int | tuple[int] | None = None,
                   color_feature_id: int | None | str = 'Iteration',
                   X_suggest: pd.DataFrame | None = None) -> Figure:
    """
    Generates a plotly figure to visualize optimization progress

    Args:
        campaign (Campaign): The campaign object containing the data.
        response_ids (list[int], optional): The indices of the responses to plot. Defaults to ``[0, 1]``.
        color_feature_id (int | None, optional): The index of the feature to use for coloring the markers.
            Defaults to ``None``, which will color by iteration.
        X_suggest (pd.DataFrame | None, optional): The suggested next experiments to evaluate.
            Defaults to ``None``.

    Returns:
        Figure: The plotly figure.

    """
    fig = go.Figure()

    if response_ids is None:
        if campaign._is_mo:
            response_ids = (0, 1)
        else:
            response_ids = (0)
    if isinstance(response_ids, int):
        response_ids = (response_ids,)

    # Extract input and output names
    out_names = []
    for id in response_ids:
        out_names.append(campaign.out.columns[id])
    X_names = list(campaign.X.columns)

    for id in response_ids:
        if id >= len(out_names):
            raise ValueError(f'Response ID {id} is out of range')
    if isinstance(color_feature_id, int):
        if color_feature_id >= len(campaign.X_space):
            raise ValueError(f'Color feature ID {color_feature_id} is out of range')
        x_color_name = X_names[color_feature_id]
    if isinstance(color_feature_id, str):
        if color_feature_id not in campaign.data.columns:
            raise ValueError(f'Color feature {color_feature_id} is not in the data')
        x_color_name = color_feature_id

    # Unpack experimental data to plot progress
    out_exp = campaign.out[out_names]
    if not campaign._is_mo:
        # In this case, we only have 1 response to plot, so use the index on x-axis
        out_exp = out_exp.reset_index(drop=False).rename(columns={'index': 'Experiment'})
        out_names.insert(0, 'Experiment')

    if color_feature_id is not None:
        x_color = campaign.data[x_color_name]
        marker_dict = dict(color=x_color,
                           colorscale=[[0, obsidian_colors.rich_blue],
                                       [0.5, obsidian_colors.teal],
                                       [1, obsidian_colors.lemon]],
                           showscale=True,
                           colorbar=dict(title=x_color_name))
    else:
        x_color = None
        marker_dict = dict(color=obsidian_colors.primary.teal)

    fig.add_trace(go.Scatter(
        x=out_exp.iloc[:, 0], y=out_exp.iloc[:, 1],
        mode='markers',
        marker=marker_dict,
        customdata=campaign.data[X_names],
        name='Data'))
    
    # Highlight the best samples
    if hasattr(campaign.optimizer, 'X_best_f_idx'):
        fig.add_trace(go.Scatter(x=pd.Series(out_exp.iloc[campaign.optimizer.X_best_f_idx, 0]),
                                 y=pd.Series(out_exp.iloc[campaign.optimizer.X_best_f_idx, 1]),
                                 mode='markers',
                                 marker=dict(symbol='diamond-open', size=14),
                                 line={'color': 'black'},
                                 legendgroup='marker_shape', showlegend=True,
                                 name='Best')
                      )
        fig.update_layout(showlegend=True)
    
    template = ["<b>"+str(param.name)+"</b>: "+" %{customdata["+str(i)+"]"
                + (":.3G}"if isinstance(param, Param_Continuous) else "}") + "<br>"
                for i, param in enumerate(campaign.X_space)]
    fig.update_traces(hovertemplate=''.join(template)
                      + '<b>' + out_names[0] + '</b>' + ": %{x:.3G}<br>"
                      + '<b>' + out_names[1] + '</b>' + ": %{y:.3G}<br>")

    if X_suggest is not None:
        if not all(x in X_suggest.columns for x in campaign.X.columns):
            raise ValueError('Suggested data must contain all responses')
        
        eval_suggest = campaign.evaluate(X_suggest)
        
        if campaign.objective is None:
            y_mu = []
            lb = []
            ub = []
            for response in out_names:
                y_mu.append(eval_suggest[response + ' (pred)'])
                lb.append(eval_suggest[response + ' lb'])
                ub.append(eval_suggest[response + ' ub'])
            error_y_plus = ub[1] - y_mu[1]
            error_y_minus = y_mu[1] - lb[1]
            error_x_plus = ub[0] - y_mu[0]
            error_x_minus = y_mu[0] - lb[0]
            
            y_mu = pd.concat(y_mu, axis=1)
            
            hovertext = out_names[0] + ' :%{x:.3G} +%{error_x.array:.2G}/- \
                        %{error_x.arrayminus:.2G}<br>' + out_names[1] + ': %{y:.3G} \
                            +%{error_y.array:.2G}/-%{error_y.arrayminus:.2G}'
            
        else:
            if campaign._is_mo:
                y_mu = eval_suggest[out_names]
            else:
                y_mu = eval_suggest[out_names[-1]]
                m_data = len(out_exp)
                m_suggest = len(X_suggest)
                y_mu = pd.concat([pd.DataFrame(np.arange(m_data, m_data+m_suggest), columns=['Experiment']),
                                  y_mu], axis=1)
            error_y_plus = error_y_minus = error_x_minus = error_x_plus = None

            hovertext = out_names[0] + ' :%{x:.3G}' + '<br>' \
                + out_names[1] + ' :%{y:.3G}'

        fig.add_trace(go.Scatter(
            x=y_mu.iloc[:, 0],
            y=y_mu.iloc[:, 1],
            mode='markers',
            marker=dict(color=obsidian_colors.accent.pastel_blue,
                        symbol='diamond-open',
                        size=7,
                        line=dict(width=2)),
            name='Suggested',
            error_y={'array': error_y_plus,
                     'arrayminus': error_y_minus,
                     'color': 'gray', 'thickness': 1},
            error_x={'array': error_x_plus,
                     'arrayminus': error_x_minus,
                     'color': 'gray', 'thickness': 1},
            hovertemplate=hovertext))
        
    fig.update_layout(
        xaxis_title=out_names[0],
        yaxis_title=out_names[1],
        title='Optimization Results'
    )

    fig.update_layout(coloraxis_colorbar=dict(yanchor="top", y=1, x=0, ticks="outside"))

    fig.update_layout(legend=dict(
        yanchor="bottom",
        y=0.05,
        xanchor="right",
        x=0.95
    ))

    fig.update_layout(width=500, height=400, template='ggplot2')
    
    return fig



from .utils import add_tab, make_input, make_dropdown, make_switch, make_slider, make_knob, make_table, make_collapse
import dash_bootstrap_components as dbc
from dash import dcc, html, Dash, dash_table, callback, Output, Input, State, ALL, MATCH

import pandas as pd

from obsidian.parameters import ParamSpace, Param_Categorical, Param_Ordinal, Param_Continuous
from obsidian.optimizer import BayesianOptimizer
from obsidian.campaign import Campaign
from obsidian.parameters import Target

from obsidian.plotting.plotly import parity_plot
from .utils import load_optimizer, center


def setup_optimize(app, app_tabs):
    
    fit_div = dbc.Container([
        dbc.Spinner(color="primary", children=[
            dbc.Button('Fit', id='button-fit', n_clicks=0, size='lg')
        ]),
        html.Br(),
        html.Br(),
        dbc.Card([
            dbc.CardHeader('Regression Statistics'),
            dbc.CardBody([
                html.Div(id='div-fit', children=[])
                ]),
            ]),
        html.Div(id='graph-parity', children=[])
        ],
        style={'textAlign': 'center'}
    )
    
    predict_div = dbc.Container([
        dbc.Spinner(color="primary", children=[
            dbc.Button('Optimize', id='button-predict', n_clicks=0, size='lg'),
            ]),
        html.Br(),
        html.Br(),
        dbc.Card([
            dbc.CardHeader('Optimal Experiments'),
            dbc.CardBody([
                html.Div(id='div-predict', children=[], style={})
                ]),
            ]),
        ],
        style={'textAlign': 'center'}
    )
    
    storage_fit = dcc.Store(id='store-fit', data=None)
    
    # candidates store
    store_candidates = dcc.Store(id='store-candidates', data={})
    
    # Suggested candidates download
    candidates_downloader = html.Div(children=[
        dbc.Button('Download Suggested Candidates', id='button-download_candidates',
                   className='me-2', color='primary'),
        dcc.Download(id='downloader-candidates')],
                                     style={'textAlign': 'center', 'margin-top': '15px'})
    
    # Add all of these elements to the app
    columns = dbc.Row([dbc.Col(fit_div, width=6), dbc.Col([predict_div, candidates_downloader], width=6)])
    elements = [html.Br(), columns, storage_fit, store_candidates]
    add_tab(app_tabs, elements, 'tab-optimize', 'Optimize')
    setup_optimize_callbacks(app)
    
    return


def setup_optimize_callbacks(app):
    
    @app.callback(
        Output('button-fit', 'n_clicks'),
        Output('store-fit', 'data'),
        Input('button-fit', 'n_clicks'),
        State('store-config', 'data'),
        State('store-X0', 'data'),
        State('store-Xspace', 'data'),
        prevent_initial_call=True
    )
    def fit_optimizer(fit_clicked, config, X0, Xspace_save):
        
        if not Xspace_save:
            return 0, None
        
        X_space = ParamSpace.load_state(Xspace_save)
        target = Target(config['response_name'], aim='max')
        campaign = Campaign(X_space, target)
        campaign.add_data(pd.DataFrame(X0))
        
        optimizer = BayesianOptimizer(X_space=campaign.X_space,
                                      surrogate=config['surrogate_params']['surrogate'])
        campaign.set_optimizer(optimizer)
        campaign.fit()
        
        return 0, campaign.optimizer.save_state()
    
    @app.callback(
        Output('div-fit', 'children'),
        Input('store-fit', 'data'),
        State('store-config', 'data'),
        State('uploader-X0', 'filename')
    )
    def fit_statistics(opt_save, config, filename):
        if opt_save is None:
            return dbc.Alert('Model must be fit first', color='info')
        
        optimizer = load_optimizer(config, opt_save)
        
        fit_stats = dbc.ListGroup(
            [
             dbc.ListGroupItem(['Model Type: ', f'{optimizer.surrogate_type}']),
             dbc.ListGroupItem(['Data Name: ', filename]),
             # TODO: Implement score for each surrogate, to support MOO
             dbc.ListGroupItem(['R', html.Sup('2'), ' Score: ', f'{optimizer.surrogate[0].r2_score: .4g}']),
             dbc.ListGroupItem(['Marginal Log Likelihood: ', f'{optimizer.surrogate[0].loss: .4g}']),
            ], flush=True
        )
        
        return fit_stats
    
    @app.callback(
        Output('graph-parity', 'children'),
        Input('store-fit', 'data'),
        State('store-config', 'data'),
        prevent_initial_call=True
    )
    def graph_parity_plot(opt_save, config):
        
        if opt_save is None:
            return None
        
        optimizer = load_optimizer(config, opt_save)
        pplot = parity_plot(optimizer)
        pplot.update_layout(height=400, width=600)
        
        graph = dcc.Graph(figure=pplot)
        
        return center(graph)
    
    @app.callback(
        Output('div-predict', 'children'),
        Output('div-predict', 'style'),
        Output('button-predict', 'n_clicks'),
        Output('store-candidates', 'data'),
        Input('button-predict', 'n_clicks'),
        State('store-config', 'data'),
        State('store-fit', 'data'),
        # prevent_initial_call = True
    )
    def predict_optimizer(predict_clicked, config, opt_save):
        
        if opt_save is None:
            if predict_clicked:
                alert_color = 'danger'
            else:
                alert_color = 'info'
            return dbc.Alert('Model must be fit first', color=alert_color), {}, predict_clicked, {}
        
        optimizer = load_optimizer(config, opt_save)
        X_suggest, eval_suggest = optimizer.suggest(**config['aq_params'])
        df_suggest = pd.concat([X_suggest, eval_suggest], axis=1)
        df_suggest.insert(loc=0, column='CandidatesID', value=df_suggest.index)
        tables = [center(make_table(df_suggest))]
        
        return tables, {'overflow-x': 'scroll'}, 0, df_suggest.to_dict()

    # Download Suggested Candidates
    @app.callback(
        Output('downloader-candidates', 'data'),
        Input('button-download_candidates', 'n_clicks'),
        State('store-candidates', 'data'),
        prevent_initial_call=True  # otherwise, the download CSV appears right away
    )
    def download_candidates(n_clicks, data):
        df = pd.DataFrame(data)
        return dcc.send_data_frame(df.to_csv, 'Suggested_Candidates.csv', index=False)
    
    return



"""Parameters that can be sampled continuously between a minimum and maximum"""

from .base import Parameter
from .utils import transform_with_type

import warnings
import numpy as np


class Param_Continuous(Parameter):
    """
    Represents a continuous parameter.

    Attributes:
        name (str): The name of the parameter.
        min (int or float): The minimum value of the parameter.
        max (int or float): The maximum value of the parameter.

    Properties:
        range (int): The range of the parameter (max - min).

        """

    @transform_with_type
    def unit_map(self, X: np.ndarray):
        """Map from measured to 0,1 space"""
        return (X-self.min)/self.range if self.range != 0 else 0*X
        
    @transform_with_type
    def unit_demap(self, X: np.ndarray):
        """Map from 0,1 to measured space"""
        return X*self.range+self.min

    encode = unit_map
    
    decode = unit_demap

    @property
    def range(self):
        """The range of the parameter (max - min)"""
        return self.max-self.min
    
    def set_search(self,
                   search_min: int | float,
                   search_max: int | float):
        """
        Set the search space for the parameter

        Args:
            search_min (int or float): The minimum value of the search space.
            search_max (int or float): The maximum value of the search space.
        """
        for val in [search_min, search_max]:
            self._validate_value(val)
            
        self.search_min = search_min
        self.search_max = search_max
    
    def open_search(self):
        """Set the search space to the parameter space"""
        self.set_search(self.min, self.max)
    
    def __repr__(self):
        """String representation of object"""
        return f"{self.__class__.__name__}(name={self.name}, min={self.min}, max={self.max})"

    def _validate_value(self, value: int | float):
        """
        Validates if the given value is within the specified range.

        Args:
            value (int | float): The value to be validated.

        Raises:
            TypeError: If the value is not a number.
            ValueError: If the value is outside of the specified range.
        """
        if not isinstance(value, (int, float)):
            raise TypeError(f'Value {value} is not a number')
        if (value < self.min) or (value > self.max):
            raise ValueError(f'Value {value} is outside of range {self.min} to {self.max}')

    def __init__(self,
                 name: str,
                 min: int | float,
                 max: int | float,
                 search_min: int | float = None,
                 search_max: int | float = None):
        super().__init__(name=name)
        if max < min:
            warnings.warn(f'Minimum value {min} is greater than maximum value {max}. Auto-swapping values.', UserWarning)
            min, max = max, min
        self.min = min
        self.max = max
        for val in [min, max]:
            self._validate_value(val)
        
        # Set the search space to the parameter space by default
        if not search_min:
            search_min = self.min
        if not search_max:
            search_max = self.max
        self.set_search(search_min, search_max)


class Param_Observational(Param_Continuous):
    """
    This is an observational numeric variable that is used for fitting but is not leveraged during optimization

    Attributes:
        name (str): The name of the parameter.
        min (int or float): The minimum value of the parameter.
        max (int or float): The maximum value of the parameter.
        design_point (int or float): The point which the observational value will be locked to during experiment optimization

    """
    def __init__(self,
                 name: str,
                 min: int | float,
                 max: int | float,
                 search_min: int | float = None,
                 search_max: int | float = None,
                 design_point: int | float | None = None):
        super().__init__(name=name, min=min, max=max, search_min=search_min, search_max=search_max)
        self.design_point = design_point = design_point if design_point is not None else max



"""
Generic simulator functions based on simple geometric surfaces.
"""

import numpy as np
import pandas as pd
import torch
import math


def paraboloid(X):
    """
    Evaluates a simple n-dimensional paraboloid at the specified location.

    Args:
        X (ndarray): An (m)-observations by (d)-features array of data to be evaluated.

    Returns:
        ndarray: An (m)-sized array of responses.

    """

    if isinstance(X, pd.DataFrame):
        X = X.to_numpy()

    d = X.shape[1]
    a = np.ones(d) * 0.5
    z = 100 - (400 / d * (X - a) ** 2).sum(axis=1)

    return z


def shifted_parab(X):
    """
    Evaluates a simple n-dimensional parabaloid at the location specified.
    Target paraboloid contains a maximum of 100 at all x=0.2

    Args:
        X (ndarray): (m)-observations by (d)-features array of data to be evaluated

    Returns:
        ndarray: (m)-sized array of responses
    """
    if isinstance(X, pd.DataFrame):
        X = X.to_numpy()

    d = X.shape[1]
    a = np.ones(d) * 0.2
    z = 100 * np.exp(-(4 / d * (X - a) ** 2).sum(axis=1))
    return z


def cornered_parab(X):
    """
    Evaluates a simple n-dimensional parabaloid at the location specified.
    
    Args:
        X (ndarray): (m)-observations by (d)-features array of data to be evaluated
    
    Returns:
        ndarray: (m)-sized array of responses
    """
    if isinstance(X, pd.DataFrame):
        X = X.to_numpy()

    d = X.shape[1]
    a = np.ones(d) * 0
    z = 100 - (400 / d * (X - a) ** 2).sum(axis=1)
    
    return z


def ackley(X):
    """
    Evaluates the N-dimensional Ackley function.

    The Ackley function is a benchmark optimization problem that is commonly used to test optimization algorithms.
    It is a multimodal function with multiple local maxima and a global minimum at (0.8, 0.8, ..., 0.8).
    This implementation of the Ackley function has been transformed to have a maximum value of 10 at the global minimum.

    Args:
        X (ndarray or DataFrame):
            An (m x d) array or DataFrame of m observations with d features to be evaluated.

    Returns:
        ndarray:
            An (m)-sized array of responses.
    """

    if isinstance(X, pd.DataFrame):
        X = X.to_numpy()

    # Rescale problem from (0,1) to desired space (-1,3)
    X = X*4-3.2

    a = 10  # default = 20
    b = 0.2  # default = 0.2
    c = 1*math.pi  # default = 2*math.pi
    d = X.shape[1]
    
    z = -a*np.exp(-b*((1/d)*np.sum(X**2, axis=1))**0.5) - np.exp((1/d)*np.sum(np.cos(c*X), axis=1))+a+np.exp(1)
    
    z = -z + 10
    
    return z


def rosenbrock(X):
    """
    Evaluates the Rosenbrock function in log10 scale, and negative transform (for max).

    The global maximum of the Rosenbrock function is at X = (1, 1, ..., 1).
    The maximum value is 10, based on the standard function minimum of 0. We have -log10(0+1e-10).

    Args:
        X (ndarray): (m)-observations by (d)-features array of data to be evaluated.

    Returns:
        ndarray: (m)-sized array of responses.
    """

    if isinstance(X, pd.DataFrame):
        X = X.to_numpy()

    # Rescale problem from (0,1) to desired space (-1,3)
    X = X*4-2

    z = np.zeros(shape=(X.shape[0]))
    x_dim = X.shape[1]
    
    b = 100
    a = 1
    
    for i in range(0, x_dim-1):
        z += b*(X[:, i+1]-X[:, i]**2)**2+(a-X[:, i])**2
    
    z = -np.log10(z+1e-10)
    
    return z


def sixhump_camel(X):
    """
    Evaluates the 6-hump camel function in 2D
    Transformed to get a max of 10 at X = (0.0898,-0.7126) and (-0.0898, 0.7126)

    Args:
        X (ndarray): (m)-observations by (d)-features array of data to be evaluated.

    Returns:
        ndarray: (m)-sized array of responses.

    """

    if X.shape[1] != 2:
        raise ValueError('Camel function only accepts 2 dimensions')

    if isinstance(X, pd.DataFrame):
        X = X.to_numpy()

    # Rescale problem from (0,1) to desired space (-2,2),(-1,1)
    x1 = X[:, 0]*4-2
    x2 = X[:, 1]*2-1
    
    z = (4-2.1*x1**2+(x1**4)/3)*x1**2+x1*x2+(-4+4*x2**2)*x2**2
    
    z = -z+1.0316+10
    
    return z


def threehump_camel(X):
    """
    Evaluates the 3-hump camel function in 2D
    Negative transform, to get max
    Two global max exist at X = (0,0)
    Max is 1

    Args:
        X (ndarray): (m)-observations by (d)-features array of data to be evaluated.

    Returns:
        ndarray: (m)-sized array of responses.

    """

    if isinstance(X, pd.DataFrame):
        X = X.to_numpy()
    
    if X.shape[1] != 2:
        raise ValueError('Camel function only accepts 2 dimensions')

    # Rescale problem from (0,1) to desired space (-2,2)
    X = X*4-2

    x1 = X[:, 0]
    x2 = X[:, 1]
    
    z = 2*x1**2-1.05*x1**4+(x1**6)/6+x1*x2+x2**2
    
    z = -z+10

    return z


def perm(X):
    """
    Evaluates the perm function in N-dimensions
    Negative transform, to get max
    Max exists at X = (1,1/2,...,1/d)
    Max is 10
    
    Args:
        X (ndarray): (m)-observations by (d)-features array of data to be evaluated.

    Returns:
        ndarray: (m)-sized array of responses.

    """
    
    if isinstance(X, pd.DataFrame):
        X = X.to_numpy()

    # Rescale problem from (0,1) to desired space (0,1.5)
    X = X*2-0.5
        
    z = np.zeros(shape=(X.shape[0]))
    x_dim = X.shape[1]
    
    beta = 10
    
    for i in range(1, x_dim+1):
        inner_sum = 0
        for j in range(1, x_dim+1):
            xj = X[:, j-1]
            inner_sum += (j+beta)*((xj**i)-(1/j)**i)
        z += inner_sum**2
    
    z = -np.log(z+1)+10
    
    return z


def branin_currin(X):
    """
    Synthetic BraninCurrin function from BoTorch test_functions

    Two objective problem composed of the Branin and Currin functions:
    
        Branin (rescaled):
    
            f_1(x) = (
            15*x_1 - 5.1 * (15 * x_0 - 5) ** 2 / (4 * pi ** 2) + 5 * (15 * x_0 - 5)
            / pi - 5
            ) ** 2 + (10 - 10 / (8 * pi)) * cos(15 * x_0 - 5))
    
        Currin:
    
            f_2(x) = (1 - exp(-1 / (2 * x_1))) * (
            2300 * x_0 ** 3 + 1900 * x_0 ** 2 + 2092 * x_0 + 60
            ) / 100 * x_0 ** 3 + 500 * x_0 ** 2 + 4 * x_0 + 20
    
    Args:
        X (ndarray): (m)-observations by (d=2)-features array of data to be evaluated.

    Returns:
        ndarray: (k=2)-sized array of responses.
    """
    if isinstance(X, pd.DataFrame):
        X = torch.tensor(X.values)
    else:
        X = torch.tensor(X)
    
    if X.shape[1] != 2:
        raise ValueError('BraninCurrin function only accepts 2 dimensions')
    
    from botorch.test_functions.multi_objective import BraninCurrin
    problem = BraninCurrin(negate=True)
    z = problem(X).numpy()
    return z


def two_leaves(X):
    """
    Simulates the two-leaf multi-output function

    Args:
        X (ndarray): (m)-observations by (d=2)-features array of data to be evaluated.

    Returns:
        ndarray: (m=2)-sized array of responses.
    """
    if isinstance(X, pd.DataFrame):
        X = torch.tensor(X.values)
    if isinstance(X, np.ndarray) and X.ndim == 1:
        X = X[np.newaxis, :]
    if isinstance(X, np.ndarray):
        X = torch.tensor(X)
    x_0 = X[..., 0]
    x_1 = X[..., 1]
    t = torch.sqrt(torch.abs((x_0-0.4)*math.pi/2)+0.1)
    y_0 = t * torch.sin(x_0*math.pi/2) * torch.pow(x_1, 2)
    y_1 = t * torch.cos(x_0*math.pi/2)  # * torch.pow(x_1,3)
    return np.array(torch.stack([y_0, y_1], dim=-1))



"""Utility functions for parameter handling"""

import numpy as np


def transform_with_type(inner):
    """
    Wraps parameter transform functions (map, encode, etc)
    in a way that returns the same type that was given
    """
    def wrapper(self, X):
        X_arr = np.array(X)
        X_t = inner(self, X_arr)
        return X_t.tolist() if isinstance(X, (list, int, float, str)) else X_t
    return wrapper



"""PyTests for obsidian.parameters"""

from obsidian.tests.param_configs import test_X_space
from obsidian.experiment import ExpDesigner
from obsidian.parameters import (
    Param_Continuous,
    Param_Observational,
    Param_Discrete_Numeric,
    Param_Categorical,
    Param_Ordinal,
    Param_Discrete,
    Task,
    ParamSpace,
    Target,
    Standard_Scaler,
    Logit_Scaler
)

from obsidian.exceptions import UnsupportedError, UnfitError
from obsidian.tests.utils import equal_state_dicts

import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal
import torch
import pytest


# Iterate over several preset parameter spaces
@pytest.fixture(params=test_X_space)
def X_space(request):
    return request.param


@pytest.fixture
def X0(X_space):
    designer = ExpDesigner(X_space, seed=0)
    X0 = designer.initialize(m_initial=20, method='LHS')
    return X0


@pytest.mark.fast
def test_param_loading(X_space):
    obj_dict = X_space.save_state()
    X_space2 = ParamSpace.load_state(obj_dict)
    for param in X_space2:
        param.__repr__()
    X_space2.__repr__()
    obj_dict2 = X_space2.save_state()
    assert equal_state_dicts(obj_dict, obj_dict2), 'Error during serialization'


@pytest.mark.fast
def test_param_mapping(X0, X_space):
    X0_map = X_space.unit_map(X0)
    X0_demap = X_space.unit_demap(X0_map)
    assert_frame_equal(X0, X0_demap, check_dtype=False)
    df_mean = X_space.mean()


@pytest.mark.fast
def test_param_encoding(X0, X_space):
    X0_encode = X_space.encode(X0)
    X0_decode = X_space.decode(X0_encode)
    assert_frame_equal(X0, X0_decode, check_dtype=False)

    # Also need to make sure that if certain parameters are excluded from the encoding, the returned df contains 0s for other categories
    designer = ExpDesigner(X_space, seed=0)
    X_zeros = designer.initialize(m_initial=10, method='Custom',
                                  sample_custom=np.zeros(shape=(10, len(X_space.params))))
    X_zeros_encode = X_space.encode(X_zeros)
    assert X_zeros_encode.shape[1] == X0_encode.shape[1]


@pytest.mark.fast
def test_param_transform_mapping(X_space):
    assert len(X_space.t_map) == X_space.n_dim
    assert len(X_space.tinv_map) == X_space.n_tdim


# Set up a sampling of parameter types for testing encoding/decoding
test_params = [
        Param_Continuous('Parameter 1', 0, 10),
        Param_Observational('Parameter 2', 0, 10),
        Param_Discrete_Numeric('Parameter 3', [-2, -1, 1, 2]),
        Param_Categorical('Parameter 4', ['A', 'B', 'C', 'D']),
        Param_Ordinal('Parameter 5', 'A, B, C, D'),
        Task('Parameter 6', ['A', 'B', 'C', 'D']),
    ]

# Set up a numbe of different data types for testing encoding/decoding
test_type = [lambda x: list(x),
             lambda x: np.array(x),
             lambda x: list(x)[0][-2],  # Single value
             ]


@pytest.mark.fast
@pytest.mark.parametrize('param, type_i', zip(test_params, test_type))
def test_param_encoding_types(param, type_i):
    
    # Set up a variety of value types to test
    cont_vals = [0, 1, 2, 3, 4]
    cat_vals = ['A', 'B', 'C', 'D']
    num_disc_vals = [-2, -1, 1, 2]

    # Make sure that 2D arrays also work!
    if isinstance(param, Param_Continuous):
        d_2 = [cont_vals] * 3
    elif isinstance(param, Param_Discrete) and not isinstance(param, Param_Discrete_Numeric):
        d_2 = [cat_vals] * 3
    elif isinstance(param, Param_Discrete_Numeric):
        d_2 = [num_disc_vals] * 3

    X = type_i(d_2)
    
    # Unit map and demap
    X_u = param.unit_map(X)
    X_u_inv = param.unit_demap(X_u)

    # Encode and decode
    X_t = param.encode(X)
    X_t_inv = param.decode(X_t)

    # Check equivalence based on type
    if isinstance(X, np.ndarray):
        assert np.all(X_u_inv == X)
        if not isinstance(param, Param_Categorical):  # Categorical params don't encode to the same shape
            assert np.all(X_t_inv == X)
    else:
        assert X_u_inv == X
        if not isinstance(param, Param_Categorical):  # Categorical params don't decode to the same shape
            assert X_t_inv == X


# VALIDATION TESTS - Force errors to be raised in object usage

numeric_list = [1, 2, 3, 4]
number = 1
string = 'A'
string_list = ['A', 'B', 'C', 'D']


@pytest.mark.fast
def test_param_indexing():
    X_space = ParamSpace(test_params)
    p0 = X_space[0]
    p0 = X_space['Parameter 1']


@pytest.mark.fast
def test_numeric_param_validation():
    # Strings for numeric
    with pytest.raises(TypeError):
        param = Param_Continuous('test', min=string, max=string)
        
    # Value outside of range
    with pytest.raises(ValueError):
        param = Param_Continuous('test', min=1, max=0)
        param._validate_value(2)
        
    # Strings for numeric
    with pytest.raises(TypeError):
        Param_Observational('test', min=string, max=string)


@pytest.mark.fast
def test_discrete_param_validation():
    # Numeric for categorical
    with pytest.raises(TypeError):
        param = Param_Categorical('test', categories=numeric_list)

    # Too long of a string
    with pytest.raises(ValueError):
        param = Param_Categorical('test', categories=string_list
                                  + ['test'*64])

    # Value not in categories
    with pytest.raises(ValueError):
        param = Param_Categorical('test', categories=string_list)
        param._validate_value('E')
        
    # Numeric for ordinal
    with pytest.raises(TypeError):
        param = Param_Ordinal('test', categories=numeric_list)
        
    # Strings for discrete numeric
    with pytest.raises(TypeError):
        param = Param_Discrete_Numeric('test', categories=string_list)
    
    # Value outside of range
    with pytest.raises(ValueError):
        param = Param_Discrete_Numeric('test', categories=numeric_list)
        param._validate_value(5)


@pytest.mark.fast
def test_paramspace_validation():
    # Overlapping namespace
    with pytest.raises(ValueError):
        X_space = ParamSpace([test_params[0], test_params[0]])
    
    # Misuse of categorical separator
    cat_sep_param = Param_Continuous('Parameter^1', 0, 10)
    with pytest.raises(ValueError):
        X_space = ParamSpace([test_params[0], cat_sep_param])
    
    # >1 Task
    with pytest.raises(UnsupportedError):
        X_space = ParamSpace([Task('Parameter X', ['A', 'B', 'C', 'D']),
                              Task('Parameter Y', ['A', 'B', 'C', 'D'])])

    test_data = pd.DataFrame(np.random.uniform(0, 1, (10, 2)), columns=['Parameter X', 'Parameter Z'])
    X_space = ParamSpace([Param_Continuous('Parameter X', min=0, max=1),
                          Param_Continuous('Parameter Y', min=0, max=1)])
    
    # Missing X names
    with pytest.raises(KeyError):
        test_encoded = X_space.encode(test_data)
        
        
@pytest.mark.fast
def test_target_validation():
    
    # Invalid aim
    with pytest.raises(ValueError):
        Target('Response1', aim='maximize')
    
    # Invalid f_transform
    with pytest.raises(KeyError):
        Target('Response1', f_transform='quadratic')
    
    test_response = torch.rand(10)
    
    # Transform before fit
    with pytest.raises(UnfitError):
        Target('Response1').transform_f(test_response)
    
    # Transform non-arraylike
    with pytest.raises(TypeError):
        Target('Response1').transform_f('ABC')
        
    # Transform non-numerical arraylike
    with pytest.raises(TypeError):
        Target('Response1').transform_f(['A', 'B', 'C'])
    
    # Transform before fit
    with pytest.raises(UnfitError):
        transform_func = Standard_Scaler()
        transform_func.forward(test_response)
    
    test_neg_response = -0.5 - torch.rand(10)
    
    # Values outside of logit range, refitting
    with pytest.warns(UserWarning):
        transform_func = Logit_Scaler(range_response=100)
        transform_func.forward(test_neg_response, fit=False)
        
        
if __name__ == '__main__':
    pytest.main([__file__, '-m', 'fast'])



"""Scalarization methods for reducing multi-output to single-output objectives"""

from .base import Objective

from obsidian.config import TORCH_DTYPE

import torch
from torch import Tensor

# Reference: https://arxiv.org/pdf/1904.05760


class Scalarization(Objective):
    """
    Base scalarization objective, which condenses multiple outputs into a single one
    
    Always a single-output objective.
    """
    
    def __init__(self) -> None:
        super().__init__(mo=False)
        
    def __repr__(self):
        """String representation of object"""
        return f'{self.__class__.__name__} (weights={self.weights.tolist()})'


class Scalar_WeightedSum(Scalarization):
    """
    Scalarizes a multi-output response using a weighted sum

    Args:
        weights (list[float]): A list of weights to be applied to the response tensor.
        
    """
    def __init__(self,
                 weights: list[float]):
        super().__init__()
        self.register_buffer('weights', torch.tensor(weights, dtype=TORCH_DTYPE))
    
    def forward(self,
                samples: Tensor,
                X: Tensor | None = None) -> Tensor:
        """
        Evaluate the objective function on the candidate set samples, X
        """
        return (self.weights * samples).sum(dim=-1)


class Scalar_WeightedNorm(Scalarization):
    """
    Scalarizes a multi-output response using a weighted norm

    Args:
        weights (list[float]): A list of weights to be applied to the response tensor.
        norm (int or None, optional): The order of vector norm to be used. If None
            is provided, the p-norm will be used
        neg (bool, optional): Whether or not to return the negative norm, which is
            required for maximizing norms based on distance to a target (utopian point).
            
    """
    def __init__(self,
                 weights: list[float],
                 norm: int | None = None,
                 neg: bool = False):
        super().__init__()
        self.register_buffer('weights', torch.tensor(weights, dtype=TORCH_DTYPE))
        
        if norm is None:
            norm = len(weights)
        self.register_buffer('norm', torch.tensor(norm, dtype=torch.long))
        
        self.register_buffer('neg', torch.tensor(neg, dtype=torch.bool))
        self.C = -1 if neg else 1

    def forward(self,
                samples: Tensor,
                X: Tensor | None = None) -> Tensor:
        """
        Evaluate the objective function on the candidate set samples, X
        """
        return self.C*(self.weights * samples).norm(self.norm, dim=-1)


class Scalar_Chebyshev(Scalarization):
    """
    Scalarizes a multi-output response using the augmented Chebyshev function.

    The augmented Chebyshev function maximizes the minimum scaled-response in conjunction with a weighted sum.

    Args:
        weights (list[float]): A list of weights to be applied to the response tensor.
        alpha (float, optional): The scaling factor for the weighted sum. Defaults to ``0.05``.
        augment (bool, optional): Flag indicating whether to perform augmentation. Defaults to ``True``.
            
    """
    def __init__(self,
                 weights: list[float],
                 alpha: float = 0.05,
                 augment: bool = True):
        super().__init__()
        self.register_buffer('weights', torch.tensor(weights, dtype=TORCH_DTYPE))
        self.register_buffer('alpha', torch.tensor(alpha, dtype=TORCH_DTYPE))
        self.register_buffer('augment', torch.tensor(augment, dtype=torch.bool))
        
    def forward(self,
                samples: Tensor,
                X: Tensor | None = None) -> Tensor:
        """
        Evaluate the objective function on the candidate set samples, X
        """
        # Augmentation seeks to maximize weighted sum
        delta = self.alpha * (self.weights * samples).sum(dim=-1) if self.augment else 0
        
        # Maximize the smallest weighted outcome
        return (-1 * self.weights * samples).max(dim=-1).values + delta



"""Parameter class definition"""

from abc import ABC, abstractmethod


class Parameter(ABC):
    """
    Base class for obsidian parameters.
    """

    def __init__(self,
                 name: str):
        self.name = name

    @abstractmethod
    def __repr__(self):
        pass  # pragma: no cover

    @abstractmethod
    def _validate_value(self,
                        value: int | float | str):
        """Validate data inputs"""
        pass  # pragma: no cover

    @abstractmethod
    def set_search(self):
        """Set the search space for the parameter"""
        pass  # pragma: no cover

    @abstractmethod
    def encode(X):
        """Encode parameter to a format that can be used for training"""
        pass  # pragma: no cover
    
    @abstractmethod
    def decode(X):
        """Decode parameter from transformed space"""
        pass  # pragma: no cover

    def save_state(self) -> dict:
        """
        Save the state of the Parameter object.

        Returns:
            dict: A dictionary containing the state of the object.
        """
        obj_dict = vars(self)
        return obj_dict

    @classmethod
    def load_state(cls,
                   obj_dict: dict):
        """
        Load the state of the Parameter object from a dictionary.

        Args:
            obj_dict (dict): A dictionary containing the state of the object.

        Returns:
            Parameter: A new instance of the Parameter class with the loaded state.
        """
        return cls(**obj_dict)


class IParamSpace(ABC):
    """
    Interface for parameter space classes.
    """

    def __init__(self, params: list[Parameter]):
        self.params = tuple(params)
    
    def __iter__(self):
        """Iterate over the parameters in the parameter space"""
        return iter(self.params)

    def __len__(self):
        """Number of parameters in the parameter space"""
        return len(self.params)
        
    def __repr__(self):
        """String representation of object"""
        return f"{self.__class__.__name__}(params={[p.name for p in self]})"

    def __getitem__(self, index: int | str) -> Parameter:
        """Retrieve a parameter by index"""
        if isinstance(index, str):
            index = self.X_names.index(index)
        return self.params[index]




"""Base class for obsidian objective functions"""

from obsidian.utils import tensordict_to_dict

from botorch.acquisition.multi_objective.objective import MCAcquisitionObjective, MCMultiOutputObjective
from abc import abstractmethod

from torch import Tensor


class Objective(MCMultiOutputObjective, MCAcquisitionObjective):
    """
    An objective function which is calculated from the model output(s) and input(s).

    Attributes:
        _is_mo (bool): A flag indicating whether the final output is multi-output.

    """

    def __init__(self,
                 mo: bool = False) -> None:

        super().__init__()
        self._is_mo = mo

    def output_shape(self,
                     samples: Tensor) -> Tensor:
        """Converts the output to the correct Torch shape based on the multi-output flag"""
        return samples.squeeze(-1) if not self._is_mo else samples

    @abstractmethod
    def forward(self,
                samples: Tensor,
                X: Tensor | None = None) -> Tensor:
        
        pass  # pragma: no cover
    
    def save_state(self) -> dict:
        """Saves the objective to a state dictionary"""
        obj_dict = {'name': self.__class__.__name__,
                    'state_dict': tensordict_to_dict(self.state_dict())}
        
        return obj_dict
    
    @classmethod
    def load_state(cls, obj_dict: dict):
        """Loads the objective from a state dictionary"""
        return cls(**obj_dict['state_dict'])

    def __repr__(self):
        """String representation of object"""
        return f'{self.__class__.__name__} (mo={self._is_mo})'



from .utils import add_tab, make_input, make_dropdown, make_switch, make_slider, make_knob, make_table, make_collapse
import dash_bootstrap_components as dbc
from dash import dcc, html, Dash, dash_table, callback, Output, Input, State, ALL, MATCH

import pandas as pd

import obsidian
from obsidian.experiment import ParamSpace
from obsidian.optimizer import BayesianOptimizer
from obsidian.plotting.plotly import parity_plot
from .utils import load_optimizer, center


def setup_infobar(app, app_infobar):
    # Infobar = 3 columns [Help, version, Contact]
    app_infobar.children = dbc.Container([
        html.Br(),
        dbc.Row([
            dbc.Col(dbc.Button('Help', outline='True', color='warning', className='me-1', id='button-help'),
                    style={'textAlign': 'left'}, width={'size': 2}),
            dbc.Col(html.Div(dbc.Badge(f'v{obsidian.__version__}', color='primary', className='me-1'), style={'textAlign': 'center'}),
                    width={'size': 2}),
            dbc.Col(dbc.Button('Contact', outline='True', color='secondary', className='me-1', id='button-contact'),
                    style={'textAlign': 'right'}, width={'size': 2}),
            ], justify='center'),
        # Pop-up for Contact Us
        dbc.Modal([
            dbc.ModalHeader(dbc.ModalTitle('Contact Us')),
            dbc.ModalBody([
                dbc.Button('Email Developers', href='mailto:kevin.stone38@gmail.com', external_link=True,
                           color='primary', className='me-1'),
                dbc.Button('Visit Our Site', href='https://msdllcpapers.github.io/obsidian/', target='_blank', external_link=True,
                           color='primary', className='me-1')
                ], style={'textAlign': 'center'}),
            dbc.ModalFooter([])
            ], id='modal-contact', is_open=False, size='xl', centered=True),
        # Pop-up for Help
        dbc.Modal([
            dbc.ModalHeader(dbc.ModalTitle('Help: APO with the obsidian Web App')),
            dbc.ModalBody('Coming soon...'),
            dbc.ModalFooter([])
            ], id='modal-help', is_open=False, size='xl', centered=True),
        html.Br()
        ], fluid=True)
    
    setup_help_callbacks(app)
    
    return


def setup_help_callbacks(app):
    
    @app.callback(
        Output('modal-help', 'is_open'),
        Input('button-help', 'n_clicks'),
        State('modal-help', 'is_open')
    )
    def modal_help(click, is_open):
        if click:
            return ~is_open
        return is_open
    
    @app.callback(
        Output('modal-contact', 'is_open'),
        Input('button-contact', 'n_clicks'),
        State('modal-contact', 'is_open')
    )
    def modal_contact(click, is_open):
        if click:
            return ~is_open
        return is_open
    
    return



"""PyTests for obsidian.constraints"""

from obsidian.campaign import Campaign
from obsidian.constraints import (
    Linear_Constraint,
    BatchVariance_Constraint,
    Blank_Constraint,
    L1_Constraint
)
from obsidian.tests.utils import DEFAULT_MOO_PATH

import pandas as pd
import pytest
import json

# Load defaults
with open(DEFAULT_MOO_PATH) as json_file:
    obj_dict = json.load(json_file)
campaign = Campaign.load_state(obj_dict)

optimizer = campaign.optimizer
X_space = campaign.X_space
target = campaign.target

test_linear = [
    Linear_Constraint(X_space, ind=[0], weights=[1], rhs=5, equality=True),
    Linear_Constraint(X_space, ind=[0, 1], weights=[1, 1], rhs=5)
]
test_nonlinear = [BatchVariance_Constraint(X_space, ind=0, tol=0.1)]
test_out = [Blank_Constraint(target), L1_Constraint(target, offset=1)]

# Run very short optimizations for testing
test_config = {'optim_samples': 2, 'optim_restarts': 2}


@pytest.mark.parametrize('out_const', test_out)
def test_out_constraints(out_const):
    out_const.__repr__()
    campaign.constrain_outputs(out_const)
    X_suggest, eval_suggest = campaign.optimizer.suggest(**test_config)
    df_suggest = pd.concat([X_suggest, eval_suggest], axis=1)
    campaign.clear_output_constraints()


@pytest.mark.parametrize('lin_const', test_linear)
def test_ineq_constraints(lin_const):
    lin_const.__repr__()
    optimizer.X_space.constrain_inputs(lin_const)
    X_suggest, eval_suggest = optimizer.suggest(**test_config)
    df_suggest = pd.concat([X_suggest, eval_suggest], axis=1)
    optimizer.X_space.clear_constraints()
    
    
@pytest.mark.parametrize('nl_const', test_nonlinear)
def test_nleq_constraints(nl_const):
    nl_const.__repr__()
    optimizer.X_space.constrain_inputs(nl_const)
    X_suggest, eval_suggest = optimizer.suggest(**test_config)
    df_suggest = pd.concat([X_suggest, eval_suggest], axis=1)
    optimizer.X_space.clear_constraints()

    
@pytest.mark.slow
def test_combo_constraints():
    X_suggest, eval_suggest = optimizer.suggest(ineq_constraints=test_linear[1],
                                                nleq_constraints=test_nonlinear[0],
                                                out_constraints=test_out[0],
                                                **test_config)
    df_suggest = pd.concat([X_suggest, eval_suggest], axis=1)
    optimizer.X_space.clear_constraints()
    
    
if __name__ == '__main__':
    pytest.main([__file__, '-m', 'not slow'])



"""obsidian: Automated experiment design and black-box optimization"""
__version__ = '0.8.6'

# Import key objects
from obsidian.campaign import Campaign
from obsidian.optimizer import BayesianOptimizer
from obsidian.surrogates import SurrogateBoTorch
from obsidian.parameters import ParamSpace, Target

# Ensure that other subpackages are imported properly for documentation
from obsidian.objectives import Objective
from obsidian.experiment import ExpDesigner
import obsidian.constraints as constraints
import obsidian.exceptions as exceptions
import obsidian.acquisition as acquisition
import obsidian.plotting as plotting



"Analysis utility functions for examining metrics over the context of an optimization campaign"

from obsidian.parameters import Param_Continuous
from obsidian.optimizer import Optimizer
import numpy as np
import pandas as pd


def calc_ofat_ranges(optimizer: Optimizer,
                     threshold: float,
                     X_ref: pd.DataFrame | pd.Series | None = None,
                     PI_range: float = 0.95,
                     steps: int = 100,
                     response_id: int = 0,
                     calc_interacts: bool = True):
    """
    Calculates an OFAT design space using confidence bounds around the optimizer prediction. Also
    includes a matrix of interaction scores.

    Args:
        optimizer (Optimizer): The optimizer object which contains a surrogate that has been fit to data
            and can be used to make predictions.
        X_ref (pd.DataFrame): The reference data point from which the OFAT variations are calculated. If ``None``,
            will use the best point from the optimizer training data.
        threshold (float): The response value threshold (minimum value) which would be considered passing for OFAT variations.
        PI_range (float, optional): The prediction interval coverage (fraction of density)
        steps (int, optional): The number of steps to use in the search for the OFAT boundaries.
            The default value is ``100``.
        response_id (int, optional): The index of the relevant response within the fitted optimizer object.
            The default value is ``0``.
        calc_interacts (bool, optional): Whether or not to return the interaction matrix; default is ``True``.

    Returns:
        ofat_ranges (pd.DataFrame): A dataframe describing the min/max OFAT values using each LB, UB, and average prediction.
            Values are scaled in the (0,1) space based on optimizer.X_space.
        cor (np.array): A matrix of interaction values between every combination of two parameters.
            Each value is the fractional reduction in size for the acceptable range envelope created by a 2-factor variation,
            in comparison to the corresponding two independent 1-factor variations. As such, diagonal elements are 0.
    """

    ofat_ranges = []
    response_name = optimizer.target[response_id].name
    
    if X_ref is None:
        X_ref = optimizer.X_best_f
    if isinstance(X_ref, pd.Series):
        X_ref = X_ref.to_frame().T

    # Calculate 1D OFAT ranges
    for p in optimizer.X_space:
        if isinstance(p, Param_Continuous):
            
            X_span = np.linspace(0, 1, steps)
            X_sim = pd.DataFrame(np.repeat(X_ref.values, repeats=steps, axis=0), columns=X_ref.columns)
            X_sim[p.name] = p.unit_demap(X_span)
            
            df_pred = optimizer.predict(X_sim, PI_range=PI_range)
            lb = df_pred[response_name + ' lb']
            ub = df_pred[response_name + ' ub']
            pred_mu = df_pred[response_name + ' (pred)']
        
            row = {'Name': p.name, 'PI Range': PI_range, 'Threshold': threshold,
                   'Aim': optimizer.target[response_id].aim, 'Response': response_name}
            
            labels = ['Mu', 'LB', 'UB']

            for label, y in zip(labels, [pred_mu, lb, ub]):
                
                if optimizer.target[response_id].aim == 'max':
                    pass_ids = np.where(y > threshold)
                else:
                    pass_ids = np.where(y < threshold)
                pass_vals = X_sim[p.name].iloc[pass_ids]

                row['Min_'+label] = p.encode(pass_vals.min())
                row['Max_'+label] = p.encode(pass_vals.max())
            ofat_ranges.append(row)

    ofat_ranges = pd.DataFrame(ofat_ranges).set_index('Name')

    # Calculate the correlation matrix as 2-FI range / diagional of 1-FI box
    if calc_interacts:
        cor = []

        # Calculate with a nested loop of parameters
        for pi in optimizer.X_space:
            cor_j = []

            if np.isnan(ofat_ranges['Min_Mu'][pi.name]):
                cor.append([np.nan]*len(optimizer.X_space))
                continue
                
            # Enumerate a grid over the passing range at the MEAN
            Xi_pass_span = pi.unit_demap(np.linspace(ofat_ranges['Min_Mu'][pi.name],
                                                     ofat_ranges['Max_Mu'][pi.name], steps))

            for pj in optimizer.X_space:
                
                if np.isnan(ofat_ranges['Min_Mu'][pj.name]):
                    cor_j.append([np.nan]*len(optimizer.X_space))
                    continue
            
                Xj_pass_span = pj.unit_demap(np.linspace(ofat_ranges['Min_Mu'][pj.name],
                                                         ofat_ranges['Max_Mu'][pj.name], steps))

                # Set up a simulation dataframe where these parameters will co-vary
                X_sim_cor = pd.DataFrame(np.repeat(X_ref.values, repeats=steps, axis=0), columns=X_ref.columns)
                X_sim_cor[pj.name] = Xj_pass_span
                X_sim_cor[pi.name] = Xi_pass_span

                # Predict the responses, and extract the target one
                pred_mu_cor_all = optimizer.predict(X_sim_cor)
                pred_mu_cor = pred_mu_cor_all.iloc[:, response_id]
                cor_passing = np.where(pred_mu_cor > threshold)[0]

                # Want to calculate the number of steps along the diagonal which pass
                # A value of 0 for cor_j means that the two parameters are independent
                if len(cor_passing) > 0:
                    start = cor_passing[0]
                    stop = cor_passing[-1]
                    pass_ij = (stop-start)/(steps-1)
                else:
                    pass_ij = 0
                cor_j.append(1 - pass_ij)

            cor.append(cor_j)
        cor = np.array(cor)
    else:
        cor = None

    return ofat_ranges, cor



"""Matplotlib figure-generating functions"""

from obsidian.optimizer import Optimizer

import matplotlib.pyplot as plt
from matplotlib.pyplot import Figure

import numpy as np
import pandas as pd


def plot_ofat_ranges(optimizer: Optimizer,
                     ofat_ranges: pd.DataFrame) -> Figure:
    """
    Plots each parameter's 1D OFAT acceptable range

    Args:
        optimizer (Optimizer): The optimizer object which contains a surrogate
            that has been fit to data and can be used to make predictions.
        ofat_ranges (pd.DataFrame): A DataFrame containing the acceptable range
            values for each parameter, at the low bound, average, and high bound.

    Returns:
        Figure: The parameter OFAT acceptable-range plot
    """

    fig = plt.figure(figsize=(2*len(ofat_ranges), 4))
    colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
    
    # Iterate over the parameteres
    for i, (p_name, row) in enumerate(ofat_ranges.iterrows()):
        color = colors[i]

        # Plot as a bar chart; x-axis is the parameter name, y-axis is the scaled value
        plt.plot([p_name, p_name], [row['Min_LB'], row['Max_LB']],
                 linewidth=6, linestyle='solid', color=color, label='High Confidence' if i == 0 else None)
        
        # If the edges of LB are too close to mean, only annotate LB (higher conf)
        if row['Min_LB'] > row['Min_Mu']:
            plt.annotate(
                f'{(optimizer.X_space[i].unit_demap(row["Min_LB"])):.2f}',
                xy=(i, row['Min_LB']), xytext=(i + 0.25, row['Min_LB']),
                fontsize=8, ha='left', va='center', rotation=0, arrowprops=dict(arrowstyle='-', color=color, lw=1))
        if row['Max_LB'] < row['Max_Mu']:
            plt.annotate(
                f'{(optimizer.X_space[i].unit_demap(row["Max_LB"])):.2f}',
                xy=(i, row['Max_LB']), xytext=(i + 0.25, row['Max_LB']),
                fontsize=8, ha='left', va='center', rotation=0, arrowprops=dict(arrowstyle='-', color=color, lw=1))

        plt.plot([p_name, p_name], [row['Min_Mu'], row['Max_Mu']], linewidth=3,
                 linestyle='solid', color=color, label='Average' if i == 0 else None)
        
        # If the edges of the mean are too close to the UB, only annotate mean (higher conf)
        plt.annotate(
            f'{(optimizer.X_space[i].unit_demap(row["Min_Mu"])):.2f}',
            xy=(i, row['Min_Mu']), xytext=(i + 0.25, row['Min_Mu']),
            fontsize=8, ha='left', va='center', rotation=0, arrowprops=dict(arrowstyle='-', color=color, lw=1))
        plt.annotate(
            f'{(optimizer.X_space[i].unit_demap(row["Max_Mu"])):.2f}',
            xy=(i, row['Max_Mu']), xytext=(i + 0.25, row['Max_Mu']),
            fontsize=8, ha='left', va='center', rotation=0, arrowprops=dict(arrowstyle='-', color=color, lw=1))
        
        # Only plot UB if it isn't already encompassed by higher-confidence ranges
        if row['Min_UB'] < row['Min_Mu']:
            plt.plot([p_name, p_name], [row['Min_UB'], row['Min_Mu']],  linewidth=1, linestyle=':', color=color)
        if row['Max_UB'] > row['Max_Mu']:
            plt.plot([p_name, p_name], [row['Max_UB'], row['Max_Mu']],  linewidth=1, linestyle=':', color=color)
        plt.plot([0], [0], linewidth=1, linestyle=':', color=color, label='Low Confidence' if i == 0 else None)
        
        # Never annotate UB (low confidence)

    alpha = ofat_ranges['PI Range'].mode().iloc[0]
    LCL = (1 - alpha) / 2
    UCL = 1 - LCL
    comparator = ">" if row['Aim'] == 'max' else "<"
    
    plt.xticks(rotation=90)
    plt.ylabel('Parameter Value (Scaled)')
    plt.ylim([-0.15, 1.15])
    plt.xlim([-1, len(ofat_ranges)])
    plt.title('Univariate Range (OFAT) Estimates from APO Model \n'
              + f'Ranges Satisfying {row["Response"]} ' + comparator + f' {row["Threshold"]} \n'
              + f'Confidence Range: {LCL*100:.1f} - {UCL*100:.1f}%',
              fontsize=10)
    plt.legend(bbox_to_anchor=(1.05, 1), loc='upper left')
    plt.close(fig)

    return fig


def plot_interactions(optimizer: Optimizer,
                      cor: np.ndarray,
                      clamp: bool = False):
    """
    Plots the parameter interaction matrix

    Args:
        optimizer (ptimizer): The optimizer object which contains a surrogate
            that has been fit to data and can be used to make predictions.
        cor (np.ndarray): The correlation matrix representing the parameter interactions.
        clamp (bool, optional): Whether to clamp the colorbar range to (0, 1).
            Defaults to ``False``.

    Returns:
        Figure: The parameter interaction plot
    """
    
    fig = plt.figure(figsize=(4, 4))
    ax = fig.gca()
    
    # Use matrix imshow to plot correlation matrix
    cax = ax.matshow(cor)
    if clamp:
        cax.set_clim(0, 1)
    
    # Set axis labels and ticks
    axis = np.arange(len(optimizer.X_space.X_names))
    names = optimizer.X_space.X_names
    ax.set_xticks(axis)
    ax.set_xticklabels(names, rotation=90)
    ax.set_yticks(axis)
    ax.set_yticklabels(names, rotation=0)
    cbar = fig.colorbar(cax)
    ax.set_title('Parameter Interactions')
    cbar.ax.set_ylabel('Range Shrinkage')
    
    # Add text annotations if correlation is greater than 0.05
    for (i, j), z in np.ndenumerate(cor):
        if z > 0.05:
            ax.text(j, i, '{:0.2f}'.format(z), ha='center', va='center', fontsize=8)
    plt.close(fig)
    
    return fig




from obsidian.exceptions import UnsupportedError

import numpy as np
from itertools import product


def factorial_DOE(d: int,
                  n_CP: int = 3,
                  shuffle: bool = True,
                  seed: int | None = None,
                  full: bool = False):
    """
    Creates a statistically designed factorial experiment (DOE).
    Specifically for 2-level designs only.
    Uses the range (0,1) for low-high instead of the typical (-1,1),
    although (-1,1) is used for calculations during alias design

    Args:
        d (int): Number of dimensions/inputs in the design.
        n_CP (int, optional): The number of centerpoints to include in the design, for estimating
            uncertainty and curvature. Default is ``3``.
        shuffle (bool, optional): Whether or not to shuffle the design or leave them in the default run
            order. Default is ``True``.
        seed (int, optional): Randomization seed. Default is ``None``.
        full (bool, optional): Whether or not to run the full DOE. Default is ``False``, which
            will lead to an efficient Res4+ design.

    Returns:
        ndarray: An (m)-by-(d) array of experiments in the (0,1) domain

    Raises:
        UnsupportedError: If the number of dimensions exceeds 12
    """
    if d > 12:
        raise UnsupportedError('The number of dimensions must be 12 or fewer for DOE (currently)')
    
    steps = np.array([-1, 1])
    CP = np.zeros(shape=(n_CP, d))
    
    # Create a dictionary that allows us to use letters
    term_codes = {}
    alphabet = 'ABCDEFGHJKLMNOPQRSTUVWXYZ'  # DESIGN EXPERT SKIPS I!!!
    for i, letter in enumerate(alphabet):
        term_codes[letter] = i
    
    res4_aliases = {2: {}, 3: {},
                    4: {'D': 'ABC'},
                    5: {'E': 'ABCD'}, 6: {'E': 'ABC', 'F': 'BCD'},
                    7: {'F': 'ABCD', 'G': 'ABCE'}, 8: {'F': 'ABC', 'G': 'ABD', 'H': 'BCDE'},
                    9: {'G': 'ABCD', 'H': 'ACEF', 'J': 'CDEF'},
                    10: {'G': 'ABCD', 'H': 'ABCE', 'J': 'ADEF', 'K': 'BDEF'},
                    11: {'G': 'ABCD', 'H': 'ABCE', 'J': 'ABDE', 'K': 'ACDEF', 'L': 'BCDEF'},
                    12: {'H': 'ABC', 'J': 'ADEF', 'K': 'BDEG', 'L': 'CDFG', 'M': 'ABCEFG'}
                    }
    
    if full:
        axes = np.tile(steps, (d, 1))
        X = np.array(list(product(*axes)))
    else:
        if res4_aliases[d] == {}:
            d_resolved = d
        else:
            decoded_keys = [term_codes[k] for k in res4_aliases[d].keys()]
            d_resolved = np.min(decoded_keys)  # e.g. If the first aliased key is G, we are resolved up to F
        axes = np.tile(steps, (d_resolved, 1))
        X = np.array(list(product(*axes)))
        for term, generator in res4_aliases[d].items():
            decoded_generator = [term_codes[g] for g in list(generator)]
            aliased_term = np.product(X[:, decoded_generator], axis=-1)[:, np.newaxis]
            X = np.hstack((X, aliased_term))
    
    # Add centerpoints then shuffle
    X = np.vstack((X, CP))
    if seed is not None:
        np.random.seed(seed)
    if shuffle:
        np.random.shuffle(X)
    # Rescale from (-1,1) to (0,0.999999)
    X = (X+1)/2 - 1e-6
    
    return X



"""Method pointers and config for parameters"""

from .continuous import (
    Param_Continuous,
    Param_Observational
)
from .discrete import (
    Param_Categorical,
    Param_Ordinal,
    Param_Discrete_Numeric,
    Task,
)

param_class_dict = {'Param_Continuous': Param_Continuous,
                    'Param_Categorical': Param_Categorical,
                    'Param_Ordinal': Param_Ordinal,
                    'Param_Discrete_Numeric': Param_Discrete_Numeric,
                    'Param_Observational': Param_Observational,
                    'Task': Task}



"""Base class for obsidian constraints"""


from abc import abstractmethod, ABC
from torch.nn import Module


class Constraint(ABC, Module):
    """
    Base class for constraints, which restrict the input or output space
    of a model or optimization problem
    """

    def __init__(self) -> None:
        super().__init__()
        return

    @abstractmethod
    def forward(self):
        pass  # pragma: no cover



"""Utility functions for surrogate model handling"""

import torch
from torch import Tensor
import torch.nn as nn
from torch.nn import Module, Parameter
import torch.optim as optim


def check_parameter_change(params: list[Parameter],
                           prev_params: list[Parameter]) -> float:
    """Check the average change in model parameters."""
    
    changes = []
    for param, prev_param in zip(params, prev_params):
        changes.append(torch.norm(param.data - prev_param.data))
    
    total_change = sum(changes)
    avg_change = total_change / len(list(params))

    return avg_change


def fit_pytorch(model: Module,
                X: Tensor,
                y: Tensor,
                loss_fcn: Module | None = None,
                verbose: bool = False,
                max_iter: int = 5000) -> None:
    """
    Fits a PyTorch model to the given input data.
    Args:
        model (Module): The PyTorch model to be trained.
        X (Tensor): The input data.
        y (Tensor): The target data.
        loss_fcn (Module, optional): The loss function to be used.
            If not provided, the Mean Squared Error (MSE) loss function will
            be used. Defaults to ``None``.
        verbose (bool, optional): Whether to print the loss at each epoch.
            Defaults to ``False``.
    """
            
    if loss_fcn is None:
        loss_fcn = nn.MSELoss()

    optimizer = optim.Adam(model.parameters(),
                           lr=1e-2,
                           weight_decay=1e-2)

    # Set up early stoppping
    avg_change = 1e10
    param_tol = 1e-2
    converge_buffer = 5
    converge_count = 0
    
    model.train()
    for epoch in range(max_iter):
  
        prev_params = [param.clone().detach() for param in model.parameters()]
        optimizer.zero_grad()
        output = model(X)
        loss = loss_fcn(output, y)
        loss.backward()
        optimizer.step()

        # Check change
        current_params = [param.clone().detach() for param in model.parameters()]
        avg_change = check_parameter_change(current_params, prev_params)

        if (epoch % 50 == 0 and verbose):
            print(f'Epoch {epoch}: Loss {loss.item()}')
        
        # If the parameter change is under tolerance, increment the counter
        if avg_change < param_tol:
            converge_count += 1
        else:
            converge_count = 0

        # Auto-converge when the count reaches the desired length
        if converge_count == converge_buffer:
            if verbose:
                print(f'Converged at epoch {epoch}')
            break

    model.eval()
    return



"""PyTests for obsidian.optimizer under single-output usage"""

from obsidian.tests.param_configs import X_sp_default, X_sp_cont_small, X_sp_cat_small

from obsidian.parameters import Target
from obsidian.experiment import ExpDesigner, Simulator
from obsidian.optimizer import BayesianOptimizer
from obsidian.experiment.benchmark import shifted_parab
from obsidian.tests.utils import approx_equal, equal_state_dicts

from botorch.models.ensemble import EnsembleModel

import pandas as pd
import numpy as np
import pytest


# Test a variety of preset parameter spaces
@pytest.fixture(params=[X_sp_default, X_sp_cont_small, X_sp_cat_small])
def X_space(request):
    return request.param


@pytest.fixture()
def Z0(X_space):
    designer = ExpDesigner(X_space, seed=1)
    X0 = designer.initialize(m_initial=len(X_space)*2, method='LHS')
    simulator = Simulator(X_space, shifted_parab, eps=0.05)
    y0 = simulator.simulate(X0)
    Z0 = pd.concat([X0, y0], axis=1)
    return Z0


@pytest.mark.fast
@pytest.mark.parametrize('f_transform', ['Standard', 'Logit_MinMax', 'Logit_Percentage', 'Identity'])
def test_f_transform(X_space, Z0, f_transform):
    # Test all of the transforms with fitting at min/max
    optimizer = BayesianOptimizer(X_space, surrogate='GP', seed=0, verbose=0)
    target = Target(name='Response', f_transform=f_transform, aim='max')
    target.__repr__()
    optimizer.fit(Z0, target=target)
    
    # Verify equivalence of f and f_inv
    y_train = optimizer.y_train
    f_train = target.transform_f(y_train)
    f_train_inv = target.transform_f(f_train, inverse=True)
    assert approx_equal(y_train.values.flatten(), f_train_inv.values.flatten())
    
    target = Target(name='Response', f_transform=f_transform, aim='min')
    optimizer.fit(Z0, target=target)

    # Verify equivalence of f and f_inv
    y_train = optimizer.y_train
    f_train = target.transform_f(y_train)
    f_train_inv = target.transform_f(f_train, inverse=True)
    assert approx_equal(y_train.values.flatten(), f_train_inv.values.flatten())


target = Target(name='Response', f_transform='Standard', aim='max')


@pytest.mark.parametrize('surrogate', [pytest.param('GP', marks=pytest.mark.fast),
                                       'GPflat',
                                       'GPprior',
                                       pytest.param('DKL', marks=pytest.mark.slow),
                                       'DNN'])
def test_optimizer_fit(X_space, surrogate, Z0, serial_test=True):
    optimizer = BayesianOptimizer(X_space, surrogate=surrogate, seed=0, verbose=3)
    
    if surrogate == 'GPflat' and not X_space.X_cont:
        # GPflat will fail will a purely categorical space because the design matrix is not p.d.
        return
    
    tol = 1e-2 if surrogate == 'DNN' else 1e-5
    
    optimizer.fit(Z0, target=target)
    if serial_test:
        obj_dict = optimizer.save_state()
        optimizer_2 = BayesianOptimizer.load_state(obj_dict)
        obj_dict2 = optimizer_2.save_state()
        assert equal_state_dicts(obj_dict, obj_dict2)
        optimizer_2.__repr__()
        y_pred = optimizer.predict(optimizer.X_train)
        y_pred_2 = optimizer_2.predict(optimizer.X_train)
        y_error = ((y_pred_2-y_pred)/y_pred.max(axis=0))

        # Only check the mean for ensemble models, as the lb/ub are not deterministic
        if isinstance(optimizer.surrogate[0].torch_model, EnsembleModel):
            check_cols = [col for col in y_pred if '(pred)' in col]
            y_error = y_error[check_cols]

        assert abs(y_error.values).max() < tol, 'Prediction error in loading parameters of saved optimizer'


# Generate a baseline optimizer to use for future tests
base_X_space = X_sp_default
optimizer = BayesianOptimizer(base_X_space, surrogate='GP', seed=0, verbose=0)
designer = ExpDesigner(base_X_space, seed=0)
X0 = designer.initialize(m_initial=6, method='LHS')
simulator = Simulator(base_X_space, shifted_parab, eps=0.05)
y0 = simulator.simulate(X0)
Z0_base = pd.concat([X0, y0], axis=1)
optimizer.fit(Z0_base, target=target)

# Run very short optimizations for testing
test_config = {'optim_samples': 2, 'optim_restarts': 2}


def test_fit_nan():
    Z0_sample = Z0_base.copy()
    for col in Z0_sample.columns:
        Z0_sample.loc[Z0_sample.sample(frac=0.1).index, col] = np.nan
    optimizer_nan = BayesianOptimizer(base_X_space, surrogate='GP', seed=0, verbose=0)
    optimizer_nan.fit(Z0_sample, target=target)


@pytest.mark.fast
def test_optimizer_pending():
    X_suggest, eval_suggest = optimizer.suggest(m_batch=2, **test_config)
    X_suggest, eval_suggest = optimizer.suggest(m_batch=1, **test_config, X_pending=X_suggest)
    X_suggest, eval_suggest = optimizer.suggest(m_batch=1, **test_config, X_pending=X_suggest, eval_pending=eval_suggest)


@pytest.mark.parametrize('m_batch', [pytest.param(1, marks=pytest.mark.fast), 3])
@pytest.mark.parametrize('fixed_var', [None, {'Parameter 1': 5}])
def test_optimizer_suggest(m_batch, fixed_var):
    X_suggest, eval_suggest = optimizer.suggest(m_batch=m_batch, fixed_var=fixed_var,
                                                acquisition=['EI', 'SF'], **test_config)
    df_suggest = pd.concat([X_suggest, eval_suggest], axis=1)


def test_suggest_searchspace():
    optimizer.X_space[0].set_search(2, 8)
    optimizer.X_space[3].set_search(['A', 'C'])
    
    X_suggest, eval_suggest = optimizer.suggest(m_batch=2, **test_config)
    df_suggest = pd.concat([X_suggest, eval_suggest], axis=1)
    
    optimizer.X_space.open_search()


test_aqs = ['NEI',
            'EI',
            {'EI': {'inflate': 0.05}},
            {'EI': {'inflate': -0.05}},
            'PI',
            {'PI': {'inflate': 0.05}},
            {'PI': {'inflate': -0.05}},
            'UCB',
            {'UCB': {'beta': 2}},
            {'UCB': {'beta': 0}},
            'NIPV',
            'SR',
            'Mean',
            'SF',
            ]


@pytest.mark.parametrize('aq', test_aqs)
def test_optimizer_aqs(aq):
    X_suggest, eval_suggest = optimizer.suggest(m_batch=2, acquisition=[aq], **test_config)
    df_suggest = pd.concat([X_suggest, eval_suggest], axis=1)


@pytest.mark.fast
def test_optimizer_maximize():
    X_suggest, eval_suggest = optimizer.maximize(**test_config)
    df_suggest = pd.concat([X_suggest, eval_suggest], axis=1)


if __name__ == '__main__':
    pytest.main([__file__, '-m', 'not slow'])



"""Method pointers and config for acquisition functions"""

from .custom import qMean, qSpaceFill

from .botorch import (
    qProbabilityOfImprovement, qUpperConfidenceBound, qSimpleRegret,
    qLogExpectedImprovement, qLogNoisyExpectedImprovement,
    qLogExpectedHypervolumeImprovement, qLogNoisyExpectedHypervolumeImprovement,
    qLogNParEGO, qNegIntegratedPosteriorVariance
)


# Available acquisition functions and method pointers
aq_class_dict = {'EI': qLogExpectedImprovement,
                 'NEI': qLogNoisyExpectedImprovement,
                 'PI': qProbabilityOfImprovement,
                 'UCB': qUpperConfidenceBound,
                 'SR': qSimpleRegret,
                 'EHVI': qLogExpectedHypervolumeImprovement,
                 'NEHVI': qLogNoisyExpectedHypervolumeImprovement,
                 'NParEGO': qLogNParEGO,
                 'NIPV': qNegIntegratedPosteriorVariance,
                 'RS': None,
                 'Mean': qMean,
                 'SF': qSpaceFill}

# Acquisition functions separated by optimization modality
universal_aqs = ['RS', 'Mean', 'SF']
valid_aqs = {'single': ['EI', 'NEI', 'PI', 'UCB', 'SR', 'NIPV'] + universal_aqs,
             'multi': ['EHVI', 'NEHVI', 'NParEGO'] + universal_aqs}
aq_defaults = {'single': 'NEI', 'multi': 'NEHVI'}


# For default values, 'optional' indicates whether or not a default value (or None) 'val' can be used
aq_hp_defaults = {
    'EI': {'inflate': {'val':  0, 'optional': True}},
    'NEI': {},
    'PI': {'inflate': {'val':  0, 'optional': True}},
    'UCB': {'beta': {'val': 1, 'optional': True}},
    'SR': {},
    'RS': {},
    'Mean': {},
    'SF': {},
    'NIPV': {},
    'EHVI': {'ref_point': {'val': None, 'optional': True}},
    'NEHVI': {'ref_point': {'val': None, 'optional': True}},
    'NParEGO': {'scalarization_weights': {'val': None, 'optional': True}},
}



"""PyTests for obsidian.campaign"""

from obsidian.campaign import Campaign
from obsidian.objectives import (
    Identity_Objective,
    Scalar_WeightedNorm,
    Scalar_WeightedSum,
    Scalar_Chebyshev,
    Feature_Objective,
    Objective_Sequence,
    Utopian_Distance,
    Index_Objective,
    Bounded_Target,
    Product_Objective,
    Divide_Objective
)

from obsidian.tests.utils import DEFAULT_MOO_PATH, equal_state_dicts
from obsidian.exceptions import IncompatibleObjectiveError

import pandas as pd
import pytest
import json

# Load default
with open(DEFAULT_MOO_PATH) as json_file:
    obj_dict = json.load(json_file)
campaign = Campaign.load_state(obj_dict)
X_space = campaign.X_space
target = campaign.target

# Run very short optimizations for testing
test_config = {'optim_samples': 2, 'optim_restarts': 2}

test_objs = [Identity_Objective(mo=len(target) > 1),
             Scalar_WeightedNorm(weights=[1, 1]),
             Feature_Objective(X_space, ind=(0,), coeff=[1]),
             Objective_Sequence([Utopian_Distance([1], target[0]), Index_Objective()]),
             Bounded_Target(bounds=[(0, 1)]*len(target), targets=target),
             Objective_Sequence([Product_Objective(ind=(0,), weights=None, const=1, new_dim=True),
                                 Divide_Objective(0, 1, new_dim=True),
                                 Divide_Objective(0, 1, new_dim=False)]),
             Product_Objective(ind=(0,), weights=[1], const=1, new_dim=False),
             None]

utopian = Utopian_Distance(utopian=[10, 10], targets=target)

test_scalars = [Scalar_WeightedSum(weights=[0.5, 0.5]),
                Objective_Sequence([utopian, Scalar_WeightedSum(weights=[0.5, 0.5])]),
                Scalar_WeightedNorm(weights=[0.5, 0.5]),
                Objective_Sequence([utopian, Scalar_WeightedNorm(weights=[0.5, 0.5], neg=True)]),
                Scalar_Chebyshev(weights=[0.5, 0.5]),
                Scalar_Chebyshev(weights=[0.5, 0.5], augment=False),
                Objective_Sequence([utopian, Scalar_Chebyshev(weights=[0.5, 0.5])])]


@pytest.mark.parametrize('obj', test_objs + test_scalars)
def test_campaign_objectives(obj):
    # Set objective, read, examine output
    campaign.set_objective(obj)
    if campaign.objective:
        campaign.objective.__repr__()
    campaign.o

    # Serialize, deserialize, re-serialize
    obj_dict = campaign.save_state()
    campaign2 = Campaign.load_state(obj_dict)
    obj_dict2 = campaign2.save_state()
    assert equal_state_dicts(obj_dict, obj_dict2), 'Error during serialization'
    
    
@pytest.mark.parametrize('m_batch', [pytest.param(1, marks=pytest.mark.fast), 2, pytest.param(5, marks=pytest.mark.slow)])
@pytest.mark.parametrize('obj', test_objs)
def test_objective_suggestions(m_batch, obj):
    optimizer = campaign.optimizer
    if obj is not None:
        if obj._is_mo and optimizer.n_response == 1 and not isinstance(obj, Feature_Objective):
            with pytest.raises(IncompatibleObjectiveError):
                X_suggest, eval_suggest = optimizer.suggest(m_batch=m_batch, objective=obj,  **test_config)
        else:
            X_suggest, eval_suggest = optimizer.suggest(m_batch=m_batch, objective=obj,  **test_config)
    else:
        X_suggest, eval_suggest = optimizer.suggest(m_batch=m_batch, **test_config)
    df_suggest = pd.concat([X_suggest, eval_suggest], axis=1)
    
 
if __name__ == '__main__':
    pytest.main([__file__, '-m', 'not slow'])



"""Explainer Class: Surrogate Model Interpretation Methods"""

from obsidian.parameters import Param_Continuous, ParamSpace
from obsidian.optimizer import Optimizer
from obsidian.exceptions import UnfitError

import shap
from shap import KernelExplainer, Explanation
from obsidian.plotting.shap import partial_dependence, one_shap_value

import numpy as np
import pandas as pd
import torch
import matplotlib.pyplot as plt
from matplotlib.figure import Figure


class Explainer():
    """
    Base class for surrogate model interpretation and post-hoc analysis.

    Properties:
        optimizer (Optimizer): obsidian Optimizer object with fitted surrogate model(s).

    Attributes:
        X_space (ParamSpace): obsidian ParamSpace object representing the allowable space for model explanation,
            could be different from the optimizer.X_space used for optimization.
        responseid (int): the index of a single outcome whose surrogate model will be explained by shap
        shap (dict): A dictionary of values containing shapley analysis results
    
    Raises:
        ValueError: If the optimizer is not fit
        
    """
    def __init__(self,
                 optimizer: Optimizer,
                 X_space: ParamSpace | None = None) -> None:
        
        if not optimizer.is_fit:
            raise UnfitError('Surrogate model in optimizer is not fit to data. ')
        
        self.set_optimizer(optimizer)
        self.X_space = optimizer.X_space if X_space is None else X_space

        self.shap = {}
    
    def __repr__(self) -> str:
        return f"Explainer(optimizer={self.optimizer})"
    
    @property
    def optimizer(self) -> Optimizer:
        """Explainer Optimizer object"""
        return self._optimizer
    
    def set_optimizer(self, optimizer: Optimizer) -> None:
        """Sets the explainer optimizer"""
        self._optimizer = optimizer
    
    def shap_explain(self,
                     responseid: int = 0,
                     n: int = 100,
                     X_ref: pd.DataFrame | None = None,
                     seed: int | None = None) -> None:
        """
        Explain the parameter sensitivities using shap values.

        Args:
            responseid (int): Index of the target response variable.
            n (int): Number of samples to generate for shap values.
            X_ref (pd.DataFrame | None): Reference DataFrame for shap values. If None,
                the mean of self.X_space will be used.
            seed (int | None): Seed value for random number generation.

        Returns:
            None: This function fits a Kernel shap explainer and save results as class attributes.

        Raises:
            ValueError: If X_ref does not contain all parameters in self.X_space or if X_ref is not a single row DataFrame.
        """
        
        self.shap['responseid'] = responseid
        
        if X_ref is None:
            X_ref = self.optimizer.X_best_f
        else:
            if not all(x in X_ref.columns for x in self.X_space.X_names):
                raise ValueError('X_ref must contain all parameters in X_space')
            if X_ref.shape[0] != 1:
                raise ValueError('X_ref must be a single row DataFrame')
        
        self.shap['X_ref'] = X_ref
        
        y_name = self.optimizer.target[responseid].name
        
        def pred_func(X):
            # helper function for shap_explain
            if isinstance(X, np.ndarray):
                X = pd.DataFrame(X, columns=self.X_space.X_names)
                for p, x in zip(self.X_space, X.columns):
                    if isinstance(p, Param_Continuous):
                        X[x] = pd.to_numeric(X[x])
            y_pred = self.optimizer.predict(X, return_f_inv=True)
            mu = y_pred[y_name+' (pred)'].values
            return mu
        
        self.shap['pred_func'] = pred_func
        self.shap['explainer'] = KernelExplainer(pred_func,  X_ref)
        self.shap['X_sample'] = self.X_space.unit_demap(
            pd.DataFrame(torch.rand(size=(n, self.X_space.n_dim)).numpy(),
                         columns=X_ref.columns)
            )
        self.shap['values'] = self.shap['explainer'].shap_values(self.shap['X_sample'],
                                                                 seed=seed, l1_reg='aic')
        self.shap['explanation'] = Explanation(self.shap['values'], feature_names=X_ref.columns)

        return

    def shap_summary(self) -> Figure:
        """SHAP Summary Plot (Beeswarm)"""
        if not self.shap:
            raise UnfitError('SHAP explainer is not fit.')
        
        fig = plt.figure()
        shap.summary_plot(self.shap['values'], self.shap['X_sample'],
                          show=False)
        plt.close(fig)
        
        return fig
    
    def shap_summary_bar(self) -> Figure:
        """SHAP Summary Plot (Bar Plot / Importance)"""
        if not self.shap:
            raise UnfitError('SHAP explainer is not fit.')
         
        fig = plt.figure()
        shap.plots.bar(self.shap['explanation'],
                       ax=fig.gca(), show=False)
        plt.close(fig)
        
        return fig

    def shap_pdp_ice(self,
                     ind: int | tuple[int] = 0,
                     ice_color_var: int | None = None,
                     hist: bool = False,
                     ace_opacity: float = 0.5,
                     npoints: int | None = None,
                     ) -> Figure:
        """
        SHAP Partial Dependence Plot with ICE
        
        Args:
            ind (int): Index of the parameter to plot
            ice_color_var (int): Index of the parameter to color the ICE lines
            hist (bool, optional): Show histogram of the feature values.
                By default ``False``
            ace_opacity (float, optional): Opacity of the ACE line. By default ``0.5``
            npoints (int, optional): Number of points for PDP x-axis. By default
                will use ``100`` for 1D PDP and ``20`` for 2D PDP.
        
        Returns:
            Matplotlib Figure of 1D or 2D PDP with ICE lines
                    
        """
        if not self.shap:
            raise UnfitError('SHAP explainer is not fit.')
        
        fig, ax = partial_dependence(
                ind=ind,
                model=self.shap['pred_func'],
                data=self.shap['X_sample'],
                ice_color_var=ice_color_var,
                hist=hist,
                ace_opacity=ace_opacity,
                show=False,
                npoints=npoints
                )
        plt.close(fig)
        
        return fig

    def shap_single_point(self,
                          X_new: pd.DataFrame | pd.Series,
                          X_ref=None) -> tuple[pd.DataFrame, Figure, Figure]:
        """
        SHAP Pair-wise Marginal Explanations
        
        Args:
            X_new (pd.DataFrame | pd.Series): New data point to explain
            X_ref (pd.DataFrame | pd.Series, optional): Reference data point
                for shap values. Default uses ``optimizer.X_best_f``
        
        Returns:
            pd.DataFrame: DataFrame containing SHAP values for the new data point
            Figure: Matplotlib Figure for SHAP values
            Figure: Matplotlib Figure for SHAP summary plot
        
        """
        if not self.shap:
            raise UnfitError('SHAP explainer is not fit.')

        if isinstance(X_new, pd.Series):
            X_new = X_new.copy().to_frame().T
        if isinstance(X_ref, pd.Series):
            X_ref = X_ref.copy().to_frame().T
            
        if not list(X_new.columns) == list(self.optimizer.X_space.X_names):
            raise ValueError('X_new must contain all parameters in X_space')
        
        if X_ref is None:
            shap_value_new = self.shap['explainer'].shap_values(X_new).squeeze()
            expected_value = self.shap['explainer'].expected_value
        else:
            if not list(X_ref.columns) == list(self.optimizer.X_space.X_names):
                raise ValueError('X_ref must contain all parameters in X_space')
            
            # if another reference point is input, need to re-fit another explainer
            explainer = shap.KernelExplainer(self.shap['pred_func'],  X_ref)
            shap_value_new = explainer.shap_values(X_new).squeeze()
            expected_value = explainer.expected_value
        
        df_shap_value_new = pd.DataFrame([shap_value_new], columns=self.X_space.X_names)
        
        fig_bar, fig_line = one_shap_value(shap_value_new, expected_value, self.X_space.X_names)
        
        return df_shap_value_new, fig_bar, fig_line

    def sensitivity(self,
                    dx: float = 1e-6,
                    X_ref: pd.DataFrame | pd.Series | None = None) -> pd.DataFrame:
        """
        Calculates the local sensitivity of the surrogate model predictions with
        respect to each parameter in the X_space.

        Args:
            optimizer (BayesianOptimizer): The optimizer object which contains a surrogate
                that has been fit to data
            and can be used to make predictions.
            dx (float, optional): The perturbation size for calculating the sensitivity.
                Defaults to ``1e-6``.
            X_ref (pd.DataFrame | pd.Series | None, optional): The reference input values for
                calculating the sensitivity.  If None, the mean of X_space will be used as the
                reference. Defaults to ``None``.

        Returns:
            pd.DataFrame: A DataFrame containing the sensitivity values for each parameter
                in X_space.

        Raises:
            ValueError: If X_ref does not contain all parameters in optimizer.X_space or if
                X_ref is not a single row DataFrame.
                
        """
        
        if isinstance(X_ref, pd.Series):
            X_ref = X_ref.copy().to_frame().T
        
        if X_ref is None:
            X_ref = self.optimizer.X_best_f
        else:
            if not all(x in X_ref.columns for x in self.optimizer.X_space.X_names):
                raise ValueError('X_ref must contain all parameters in X_space')
            if X_ref.shape[0] != 1:
                raise ValueError('X_ref must be a single row DataFrame')
        
        y_ref = self.optimizer.predict(X_ref)
        
        sens = {}
        
        # Only do positive perturbation, for simplicity
        for param in self.optimizer.X_space:
            base = param.unit_map(X_ref[param.name].values)[0]
            # Space already mapped to (0,1), use absolute perturbation
            dx_pos = np.array(base+dx).reshape(-1, 1)
            X_sim = X_ref.copy()
            X_sim[param.name] = param.unit_demap(dx_pos)[0]
            y_sim = self.optimizer.predict(X_sim)
            dydx = (y_sim - y_ref)/dx
            sens[param.name] = dydx.to_dict('records')[0]
        
        df_sens = pd.DataFrame(sens).T[[y+' (pred)' for y in self.optimizer.y_names]]
        
        return df_sens



"""Simulate virtual experimental data"""

from obsidian.parameters import ParamSpace

from typing import Callable
import pandas as pd
import numpy as np
import warnings


class Simulator:
    """
    Simulator class for generating in-silico responses to requested experiments.

    This class provides functionality to simulate responses to a set of experiments based on a given response function.
    The simulated responses can be subject to error, which is controlled by the `eps` parameter.

    Attributes:
        X_space (ParamSpace): The ParamSpace object representing the allowable space for optimization.
        response_function (Callable): The callable function used to convert experiments to responses.
        name (str or list[str]): Name of the simulated output(s). Default is ``Response``.
        eps (float or list[float]): The simulated error to apply, as the standard deviation of the Standard
            Normal distribution. Default is ``0``.
        kwargs (dict): Optional hyperparameters for the response function.

    Raises:
        TypeError: If response_function is not a callable function.
        TypeError: If X_space is not an obsidian ParamSpace object.
        
    """

    def __init__(self,
                 X_space: ParamSpace,
                 response_function: Callable,
                 name: str | list[str] = 'Response',
                 eps: float | list[float] = 0.0,
                 **kwargs):
        
        if not callable(response_function):
            raise TypeError('Response generator must be a callable function')
        if not isinstance(X_space, ParamSpace):
            raise TypeError('X_space must be an obsidian ParamSpace object')
        
        self.X_space = X_space
        self.response_function = response_function
        self.name = name
        self.eps = eps if isinstance(eps, list) else [eps]
        self.kwargs = kwargs
    
    def __repr__(self):
        """String representation of object"""
        return f" obsidian Simulator(response_function={self.response_function.__name__}, eps={self.eps})"

    def simulate(self,
                 X_prop: pd.DataFrame) -> np.ndarray:
        """
        Generates a response to a set of experiments.

        Currently, response function only handles strictly numeric values and categories are manually penalized.

        Args:
            X_prop (pd.DataFrame): Proposed experiments to evaluate.


        Returns:
            np.ndarray: Array of response values to experiments.
        """
        # De-map everything into 0,1 based on ranges
        X = self.X_space.unit_map(X_prop).values
        
        y_sim = self.response_function(X)
        
        # Apply error
        # Expand length of eps to match number of outputs
        if len(self.eps) == 1:
            self.eps *= y_sim.ndim
        if y_sim.ndim == 1:
            y_sim = y_sim.reshape(-1, 1)
        for i in range(y_sim.shape[1]):
            rel_error = np.random.normal(loc=1, scale=self.eps[i], size=y_sim.shape[0])
            y_sim[:, i] *= rel_error

        # Handle naming conventions
        y_dims = y_sim.shape[1]
        if isinstance(self.name, list):
            if len(self.name) != y_dims:
                warnings.warn("Number of names does not match the number of dimensions. Using default response names.")
                self.name = [f'{self.name} {i+1}' for i in range(y_dims)]
        else:
            if y_dims == 1:
                self.name = [self.name]
            else:
                self.name = [f'{self.name} {i+1}' for i in range(y_dims)]

        df_sim = pd.DataFrame(y_sim, columns=self.name)

        return df_sim



"""Utility functions"""

from obsidian.parameters import Target
from torch import Tensor
import torch


def unscale_samples(samples: Tensor,
                    target: list[Target]) -> Tensor:
    """
    Unscale the scaled samples based on the given target(s).

    Args:
        samples (Tensor): The scaled samples to be unscaled.
        target (Target | list[Target]): The target(s) used for unscaling.
            It can be a single Target object or a list of Target objects.

    Returns:
        Tensor: The unscaled samples.

    Raises:
        TypeError: If the target is not a Target object or a list of Target objects.
        ValueError: If the number of constraint targets does not match the number of output dimensions.

    """

    # samples = sample_shape x batch_shape (x q) x m
    shape = samples.shape
    if shape[-1] != len(target):
        raise ValueError('Number of constraint targets must match number of output dimensions')
    for i, t in enumerate(target):
        samples[..., i] = Tensor(t.transform_f(samples[..., i].flatten().detach(),
                                               inverse=True).values).reshape(shape[:-1])
    return samples


def tensordict_to_dict(state_dict: dict[Tensor]) -> dict[float | int | str]:
    """
    Converts a dictionary of tensors to a dictionary of numpy arrays or string representations.

    Args:
        state_dict (dict): A dictionary containing tensors.

    Returns:
        dict: A dictionary with the same keys as `state_dict`, but with values converted to numpy arrays or string representations.

    """
    dict = {}
    for param in state_dict:
        # Convert to data first, otherwise boolean comparisons will not work
        dict[param] = state_dict[param].cpu().data.numpy().tolist()
        if dict[param] == torch.inf:
            dict[param] = 'inf'
        elif dict[param] == -torch.inf:
            dict[param] = '-inf'
    return dict


def dict_to_tensordict(dict: dict[float | int | str]) -> dict[Tensor]:
    """
    Converts a dictionary of values to a dictionary of tensors.

    Args:
        dict (dict): The input dictionary.

    Returns:
        dict: A dictionary where the values are converted to tensors.
    """
    state_dict = {}
    # Make sure all parameters are tensors
    for param in dict:
        if dict[param] == 'inf':
            state_dict[param] = torch.tensor(torch.inf)
        elif dict[param] == '-inf':
            state_dict[param] = torch.tensor(-torch.inf)
        else:
            state_dict[param] = torch.tensor(dict[param])
    return state_dict



"""Custom implementations of Gaussian Process models using BoTorch API"""

from obsidian.config import TORCH_DTYPE

from botorch.models.gpytorch import GPyTorchModel
from gpytorch.distributions import MultivariateNormal
from gpytorch.likelihoods import GaussianLikelihood

from gpytorch.means import ConstantMean
from gpytorch.models import ExactGP

from gpytorch.kernels import ScaleKernel, MaternKernel
from gpytorch.priors import GammaPrior
from gpytorch.constraints import GreaterThan

import torch
from torch import nn
import gpytorch


class PriorGP(ExactGP, GPyTorchModel):
    """
    Class which builds a GP with custom prior distributions;
    by default set to the values of BoTorch SingleTaskGP
    """
    _num_outputs = 1

    def __init__(self, train_X, train_Y):

        # Prior on the noise (custom likelihood)
        noise_prior = GammaPrior(1.1, 0.05)
        noise_prior_mode = (noise_prior.concentration - 1) / noise_prior.rate
        likelihood = GaussianLikelihood(
                noise_prior=noise_prior,
                noise_constraint=GreaterThan(
                    1e-4,
                    transform=None,
                    initial_value=noise_prior_mode,
                ),
            )
        
        super().__init__(train_X, train_Y.squeeze(-1), likelihood)
        
        n_dims = train_X.shape[-1]
        
        self.mean_module = ConstantMean()
        self.covar_module = ScaleKernel(
            base_kernel=MaternKernel(nu=2.5, ard_num_dims=n_dims,
                                     lengthscale_prior=GammaPrior(3.0, 6.0)),
            outputscale_prior=GammaPrior(2.0, 0.15),
        )
        
    def forward(self, x):
        """Evaluate the forward pass of the model on inputs X"""
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultivariateNormal(mean_x, covar_x)


class FlatGP(ExactGP, GPyTorchModel):
    """
    GP Surrogate with non-informative or no prior distributions
    """
    _num_outputs = 1

    def __init__(self, train_X, train_Y, nu=2.5):

        super().__init__(train_X, train_Y.squeeze(-1), GaussianLikelihood())
        
        n_dims = train_X.shape[-1]
        
        self.mean_module = ConstantMean()
        self.covar_module = ScaleKernel(
            base_kernel=MaternKernel(nu=nu, ard_num_dims=n_dims),
        )
        
    def forward(self, x):
        """Evaluate the forward pass of the model on inputs X"""
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultivariateNormal(mean_x, covar_x)


class DKLGP(ExactGP, GPyTorchModel):
    """
    GP surrogate with a FF NN feature extractor
    """
    _num_outputs = 1

    def __init__(self, train_X, train_Y):

        super().__init__(train_X, train_Y.squeeze(-1), GaussianLikelihood())
        
        n_dims = train_X.shape[-1]
        
        self.mean_module = ConstantMean()
        self.covar_module = ScaleKernel(
            base_kernel=MaternKernel(nu=2.5, ard_num_dims=n_dims),
        )
        
        # Set up the NN feature extractor
        torch.set_default_dtype(TORCH_DTYPE)
        self.feature_extractor = nn.Sequential(
            # Hiden layer 1
            nn.Linear(n_dims, n_dims),
            nn.PReLU(),
            # Hidden layer 2
            nn.Linear(n_dims, n_dims),
            nn.PReLU(),
            # #Output layer
            nn.Linear(n_dims, n_dims)
            )
        
        # Use a scaler to make sure the GP only sees well-conditioned values
        self.scale_to_bounds = gpytorch.utils.grid.ScaleToBounds(0, 1)

    def forward(self, x):
        """Evaluate the forward pass of the model on inputs X"""
        # Pass the data through the feature extractor and scaler
        projected_x = self.feature_extractor(x)
        projected_x = self.scale_to_bounds(projected_x)
        
        mean_x = self.mean_module(projected_x)
        covar_x = self.covar_module(projected_x)
        return MultivariateNormal(mean_x, covar_x)



"""Design initial experiments"""

from .utils import factorial_DOE

from obsidian.parameters import ParamSpace
from obsidian.exceptions import UnsupportedError

from botorch.utils.sampling import draw_sobol_samples
from numpy.typing import ArrayLike
from scipy.stats import qmc

import torch
from torch import Tensor
import pandas as pd
import warnings


class ExpDesigner:
    """
    ExpDesigner is a base class for designing experiments in a parameter space.

    Attributes:
        X_space (ParamSpace): The parameter space for the experiment.
        seed (int | None): The randomization seed.
    
    Raises:
        TypeError: If X_space is not an obsidian ParamSpace object.
    """

    def __init__(self,
                 X_space: ParamSpace,
                 seed: int | None = None):
        if not isinstance(X_space, ParamSpace):
            raise TypeError('X_space must be an obsidian ParamSpace object')
        
        self.X_space = X_space
        self.seed = seed

    def __repr__(self):
        """String representation of object"""
        return f"obsidian ExpDesigner(X_space={self.X_space})"

    def initialize(self,
                   m_initial: int | None = None,
                   method: str = 'LHS',
                   sample_custom: Tensor | ArrayLike | None = None) -> pd.DataFrame:
        """
        Initializes the experiment design.

        Args:
            m_initial (int): The number of experiments to initialize.
            method (str, optional): The method to use for initialization. Defaults to ``'LHS'``.
            seed (int | None, optional): The randomization seed. Defaults to ``None``.
            sample_custom (Tensor | ArrayLike | None, optional): Custom samples for initialization. Defaults to ``None``.

        Returns:
            pd.DataFrame: The initialized experiment design.

        Raises:
            KeyError: If method is not one of the supported methods.
            ValueError: If sample_custom is None when method is 'Custom'.
            ValueError: If the number of columns in sample_custom does not match the size of the feature space.
        """
        d = self.X_space.n_dim

        if m_initial is None:
            m_initial = int(d*2)
        m = m_initial
        seed = self.seed

        method_dict = {
            'LHS': lambda d, m: torch.tensor(
                qmc.LatinHypercube(d=d, scramble=False, seed=seed, strength=1, optimization='random-cd').random(n=m)),
            'Random': lambda d, m: torch.rand(size=(m, d)),
            'Sobol': lambda d, m: draw_sobol_samples(
                bounds=torch.tensor([0.0, 1.0]).reshape(2, 1).repeat(1, d), n=m, q=1).squeeze(1),
            'Custom': lambda d, m: torch.tensor(sample_custom),
            'DOE_full': lambda d, m: torch.tensor(factorial_DOE(d=d, n_CP=3, shuffle=True, seed=seed, full=True)),
            'DOE_res4': lambda d, m: torch.tensor(factorial_DOE(d=d, n_CP=3, shuffle=True, seed=seed))
        }
        
        if method not in method_dict.keys():
            raise KeyError(f'Method must be one of {method_dict.keys()}')
        if method == 'Custom':
            if sample_custom is None:
                raise ValueError('Must provide samples for custom')
        if method in ['DOE_full', 'DOE_res4']:
            if self.X_space.X_discrete:
                raise UnsupportedError('DOE methods not currently designed for discrete parameters')

        if seed is not None:
            torch.manual_seed(seed)
            torch.use_deterministic_algorithms(True)
            
        if sample_custom is not None:
            if sample_custom.shape[1] != d:
                raise ValueError('Columns in custom sample do not match size of feature space')

        # Generate [0-1) samples for each parameter
        sample = method_dict[method](d, m)

        m_required = sample.shape[0]
        
        if m_required > m:
            warnings.warn(f'The number of experiments required to initialize the requested design \
                          ({m_required}) exceeds the m_initial specified ({m}). \
                            Proceeding with larger number of experiments.')
        elif m_required < m:
            print(f'The number of initialization experiments ({m}) exceeds the required \
                   number for the requested design ({m_required}). Filling with randomized experiments.')
            excess = m - m_required
            sample_add = torch.rand(size=(excess, d))
            sample = torch.vstack((sample, sample_add))

        sample = pd.DataFrame(sample.numpy(), columns=self.X_space.X_names)
        
        # Reset parameters to 0 which are not allowed to vary in X_space
        for param in self.X_space.X_static:
            sample[param] = 0
            
        # Map samples into parameter space
        X_0 = self.X_space.unit_demap(sample)
                
        return X_0
